from pydantic import BaseModel
from enum import Enum
from typing import Dict, Any, Type, Optional, List

GSI_VALUE = "ok"
GSI_FIELD_NAME = "gsi_status"

try:
    from ProvenaInterfaces.ProvenanceModels import ModelRunRecord
    from ProvenaInterfaces.ProvenanceAPI import ProvenanceRecordInfo
    from ProvenaInterfaces.RegistryModels import ItemSubType
    from ProvenaInterfaces.SharedTypes import UserInfo, EncryptedUserInfo
except:
    from .ProvenanceModels import ModelRunRecord
    from .ProvenanceAPI import ProvenanceRecordInfo
    from .RegistryModels import ItemSubType
    from .SharedTypes import UserInfo, EncryptedUserInfo


class JobType(str, Enum):
    PROV_LODGE = "PROV_LODGE"
    REGISTRY = "REGISTRY"
    EMAIL = "EMAIL"


class JobSubType(str, Enum):
    # PROV LODGE
    PROV_LODGE_WAKE_UP = "PROV_LODGE_WAKE_UP"
    MODEL_RUN_PROV_LODGE = "MODEL_RUN_PROV_LODGE"
    MODEL_RUN_LODGE_ONLY = "MODEL_RUN_LODGE_ONLY"
    MODEL_RUN_BATCH_SUBMIT = "MODEL_RUN_BATCH_SUBMIT"
    LODGE_CREATE_ACTIVITY = "LODGE_CREATE_ACTIVITY"
    LODGE_VERSION_ACTIVITY = "LODGE_VERSION_ACTIVITY"
    MODEL_RUN_UPDATE = "MODEL_RUN_UPDATE"
    MODEL_RUN_UPDATE_LODGE_ONLY = "MODEL_RUN_UPDATE_LODGE_ONLY"

    # REGISTRY
    REGISTRY_WAKE_UP = "REGISTRY_WAKE_UP"
    REGISTER_CREATE_ACTIVITY = "REGISTER_CREATE_ACTIVITY"
    REGISTER_VERSION_ACTIVITY = "REGISTER_VERSION_ACTIVITY"

    # EMAIL
    EMAIL_WAKE_UP = "EMAIL_WAKE_UP"
    SEND_EMAIL = "SEND_EMAIL"


class JobStatus(str, Enum):
    PENDING = "PENDING"
    DEQUEUED = "DEQUEUED"
    IN_PROGRESS = "IN_PROGRESS"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"


class JobTableBase(BaseModel):
    # partition key
    session_id: str

    # timestamp (unix timestamp)
    # sort key
    created_timestamp: int

    # who owns this job?
    username: str

    # batch id - optional for batches
    # GSI on batch id
    batch_id: Optional[str] = None

    # this contains the specific job info
    payload: Dict[str, Any]

    # job type
    job_type: JobType
    # job sub type
    job_sub_type: JobSubType

    # universal gsi field
    gsi_status: str = GSI_VALUE


# SNS Topic payload
class JobSnsPayload(JobTableBase):
    pass

# Job Status Table Entry


class JobStatusTable(JobTableBase):
    # What is the status of the job?
    status: JobStatus

    # Contains any error info if relevant to status
    info: Optional[str] = None

    # If the result has completed successfully, payload associated with it
    result: Optional[Dict[str, Any]] = None

# =====================
# JOB SUB TYPE PAYLOADS
# =====================

# All wake up models are the same


class WakeUpPayload(BaseModel):
    reason: Optional[str] = None


class WakeUpResult(BaseModel):
    pass

# PROV LODGE
# ----------

# PROV LODGE MODEL RUN


class ProvLodgeModelRunPayload(BaseModel):
    record: ModelRunRecord
    revalidate: bool
    # this includes encrypted user context information
    user_info: EncryptedUserInfo


class ProvLodgeModelRunResult(BaseModel):
    record: ProvenanceRecordInfo

# PROV UPDATE MODEL RUN


class ProvLodgeUpdatePayload(BaseModel):
    model_run_record_id: str
    updated_record: ModelRunRecord
    reason: str
    revalidate: bool
    # this includes encrypted user context information
    user_info: EncryptedUserInfo


class ProvLodgeUpdateResult(BaseModel):
    # The updated record
    record: ProvenanceRecordInfo

# PROV UPDATE LODGE ONLY


class ProvLodgeUpdateLodgeOnlyPayload(BaseModel):
    model_run_record_id: str
    updated_record: ModelRunRecord
    revalidate: bool
    user_info: EncryptedUserInfo


class ProvLodgeUpdateLodgeOnlyResult(BaseModel):
    pass

# PROV LODGE ONLY MODEL RUN


class ProvLodgeModelRunLodgeOnlyPayload(BaseModel):
    model_run_record_id: str
    record: ModelRunRecord
    revalidate: bool
    user_info: EncryptedUserInfo


class ProvLodgeModelRunLodgeOnlyResult(BaseModel):
    record: ProvenanceRecordInfo


# PROV LODGE BATCH SUBMIT


class ProvLodgeBatchSubmitPayload(BaseModel):
    # The list of records to lodge
    records: List[ModelRunRecord]
    # this includes encrypted user context information
    user_info: EncryptedUserInfo


class ProvLodgeBatchSubmitResult(BaseModel):
    # The batch which contains spin off jobs
    batch_id: str

# PROV LODGE CREATION


class ProvLodgeCreationPayload(BaseModel):
    # needs to know the created item ID and the activity ID
    created_item_id: str
    creation_activity_id: str
    # easier to resolve this in registry context for now
    linked_person_id: str
    created_item_subtype: ItemSubType


class ProvLodgeCreationResult(BaseModel):
    # No meaningful response
    pass

# PROV LODGE VERSION


class ProvLodgeVersionPayload(BaseModel):
    # needs to know the created item ID and the activity ID
    from_version_id: str
    to_version_id: str
    version_activity_id: str
    linked_person_id: str
    item_subtype: ItemSubType


class ProvLodgeVersionResult(BaseModel):
    # No meaningful response
    pass

# REGISTRY
# --------

# REGISTER CREATE ACTIVITY


class RegistryRegisterCreateActivityPayload(BaseModel):
    created_item_id: str
    created_item_subtype: ItemSubType
    # easier to resolve this in registry context for now
    linked_person_id: str


class RegistryRegisterCreateActivityResult(BaseModel):
    creation_activity_id: str
    lodge_session_id: str

# REGISTER VERSION ACTIVITY


class RegistryRegisterVersionActivityPayload(BaseModel):
    # Why was this version performed
    reason: str

    # New version number
    version_number: int

    # FROM
    from_version_id: str
    # TO
    to_version_id: str
    # WHO
    linked_person_id: str

    # Which subtype?
    item_subtype: ItemSubType


class RegistryRegisterVersionActivityResult(BaseModel):
    version_activity_id: str
    lodge_session_id: str


# EMAIL
# -----

# SEND EMAIL

class EmailSendEmailPayload(BaseModel):
    email_to: str
    subject: str
    body: str
    reason: str


class EmailSendEmailResult(BaseModel):
    # No meaningful payload in result - error reported if failed
    pass


# ====================
# JOB TYPE PAYLOAD MAP
# ====================
JOB_TYPE_PAYLOAD_MAP: Dict[JobSubType, Type[BaseModel]] = {
    # Prov Lodge
    JobSubType.PROV_LODGE_WAKE_UP: WakeUpPayload,
    JobSubType.MODEL_RUN_PROV_LODGE: ProvLodgeModelRunPayload,
    JobSubType.MODEL_RUN_LODGE_ONLY: ProvLodgeModelRunLodgeOnlyPayload,
    JobSubType.MODEL_RUN_BATCH_SUBMIT: ProvLodgeBatchSubmitPayload,
    JobSubType.LODGE_CREATE_ACTIVITY: ProvLodgeCreationPayload,
    JobSubType.LODGE_VERSION_ACTIVITY: ProvLodgeVersionPayload,
    JobSubType.MODEL_RUN_UPDATE: ProvLodgeUpdatePayload,
    JobSubType.MODEL_RUN_UPDATE_LODGE_ONLY: ProvLodgeUpdateLodgeOnlyPayload,

    # Registry
    JobSubType.REGISTRY_WAKE_UP: WakeUpPayload,
    JobSubType.REGISTER_CREATE_ACTIVITY: RegistryRegisterCreateActivityPayload,
    JobSubType.REGISTER_VERSION_ACTIVITY: RegistryRegisterVersionActivityPayload,

    # Email
    JobSubType.EMAIL_WAKE_UP: WakeUpPayload,
    JobSubType.SEND_EMAIL: EmailSendEmailPayload,
}

# ===================
# JOB TYPE RESULT MAP
# ===================

JOB_TYPE_RESULT_MAP: Dict[JobSubType, Type[BaseModel]] = {
    # Prov Lodge
    JobSubType.PROV_LODGE_WAKE_UP: WakeUpResult,
    JobSubType.MODEL_RUN_PROV_LODGE: ProvLodgeModelRunResult,
    JobSubType.MODEL_RUN_LODGE_ONLY: ProvLodgeModelRunLodgeOnlyResult,
    JobSubType.MODEL_RUN_BATCH_SUBMIT: ProvLodgeBatchSubmitResult,
    JobSubType.LODGE_CREATE_ACTIVITY: ProvLodgeCreationResult,
    JobSubType.LODGE_VERSION_ACTIVITY: ProvLodgeVersionResult,
    JobSubType.MODEL_RUN_UPDATE: ProvLodgeUpdateResult,
    JobSubType.MODEL_RUN_UPDATE_LODGE_ONLY: ProvLodgeUpdateLodgeOnlyResult,

    # Registry
    JobSubType.REGISTRY_WAKE_UP: WakeUpResult,
    JobSubType.REGISTER_CREATE_ACTIVITY: RegistryRegisterCreateActivityResult,
    JobSubType.REGISTER_VERSION_ACTIVITY: RegistryRegisterVersionActivityResult,

    # Email
    JobSubType.EMAIL_WAKE_UP: WakeUpResult,
    JobSubType.SEND_EMAIL: EmailSendEmailResult,
}



from pydantic import BaseModel
from typing_extensions import Self
try:
    from ProvenaInterfaces.RegistryModels import *
    from ProvenaInterfaces.SharedTypes import StatusResponse, Status, ImportMode
except:
    from .RegistryModels import *
    from .SharedTypes import Status, StatusResponse, ImportMode

from typing import Any, Dict, Union, List

'''
LOCK
'''


class LockHistoryResponse(StatusResponse):
    history: Optional[List[LockEvent]] = None


class LockChangeRequest(BaseModel):
    id: str
    reason: str


class LockStatusResponse(BaseModel):
    locked: bool


"""
===============
VERSION REVERT
===============
"""


class ItemRevertRequest(BaseModel):
    # id of the item to revert
    id: str
    # id of the history entry for the item
    history_id: int
    # the reason for this revert
    reason: str


class ItemRevertResponse(StatusResponse):
    pass


"""
==============
AUTH RESPONSES
==============
"""


class DescribeAccessResponse(BaseModel):
    # combines the status and access description responses
    roles: Roles


class AuthRolesResponse(BaseModel):
    # the list of available roles for a given type
    roles: List[DescribedRole]

# Shared models


class QueryRecordTypes(str, Enum):
    # All types are returned - no filters
    ALL = "ALL"
    # Only seed items are returned
    SEED_ONLY = "SEED_ONLY"
    # Only complete items are returned (DEFAULT)
    COMPLETE_ONLY = "COMPLETE_ONLY"


class QueryDatasetReleaseStatusType(str, Enum):
    # should mirror ReleaseStatus in RegistryModels.py
    NOT_RELEASED = "NOT_RELEASED"
    PENDING = "PENDING"
    RELEASED = "RELEASED"


class QueryFilter(BaseModel):
    item_category: Optional[ItemCategory] = None
    item_subtype: Optional[ItemSubType] = None
    # Defaults to complete only
    record_type: QueryRecordTypes = QueryRecordTypes.COMPLETE_ONLY


class SortType(str, Enum):
    CREATED_TIME = "CREATED_TIME"
    UPDATED_TIME = "UPDATED_TIME"
    DISPLAY_NAME = "DISPLAY_NAME"
    RELEASE_TIMESTAMP = "RELEASE_TIMESTAMP"
    ACCESS_INFO_URI_BEGINS_WITH = "ACCESS_INFO_URI_BEGINS_WITH"


# newest first (assuming timestamp)
DEFAULT_SORTING_ASCENDING = False


class SortOptions(BaseModel):
    sort_type: Optional[SortType] = None
    ascending: bool = DEFAULT_SORTING_ASCENDING
    begins_with: Optional[str] = None

    # validate begins_with is only used for certain sort types
    @model_validator(mode='before')
    @classmethod
    def validate_sort_options(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        # Handle case where values might be a model instance instead of dict
        if not isinstance(values, dict):
            return values

        # not all sort keys are 'begins_with' compatible
        valid_begins_with_sort_types = [
            SortType.ACCESS_INFO_URI_BEGINS_WITH.name]

        if values.get("begins_with"):
            if not values.get("sort_type"):
                raise ValueError(
                    "Cannot filter by begins_with without specifying sort_type")
            if values.get("sort_type") not in valid_begins_with_sort_types:
                raise ValueError(
                    f"Cannot filter by begins_with without specifying a valid sort_type. Must be one of {valid_begins_with_sort_types}")

        # validate if not begins_with, then sort_type must be not in *valid*begins_with_sort_types
        if not values.get("begins_with") and values.get("sort_type") in valid_begins_with_sort_types:
            raise ValueError(
                f"Cannot filter by begins-with sort_type {values.get('sort_type')} without specifying a begins_with value")

        return values


class SubtypeFilterOptions(BaseModel):
    # all filters that are not subtype specific
    record_type: QueryRecordTypes = QueryRecordTypes.COMPLETE_ONLY


class FilterOptions(SubtypeFilterOptions):
    item_subtype: Optional[ItemSubType] = None
    # release_status: Optional[ReleasedStatus] # changing this to be a post filter
    release_reviewer: Optional[IdentifiedResource] = None
    release_status: Optional[QueryDatasetReleaseStatusType] = None

    # validate not having both values present
    @model_validator(mode='after')
    def validate_filter_options(self) -> Self:
        if self.item_subtype and self.release_status:
            raise ValueError(
                "Cannot filter by both item_subtype and release_status")
        return self


class FilterType(str, Enum):
    ITEM_SUBTYPE = "ITEM_SUBTYPE"
    RELEASE_REVIEWER = "RELEASE_REVIEWER"
    # RELEASE_STATUS = "RELEASE_STATUS"


PaginationKey = Dict[str, Any]

DEFAULT_PAGE_SIZE = 20


class SubtypeListRequest(BaseModel):
    filter_by: Optional[SubtypeFilterOptions] = None
    sort_by: Optional[SortOptions] = None
    pagination_key: Optional[PaginationKey] = None
    page_size: int = DEFAULT_PAGE_SIZE


class NoFilterSubtypeListRequest(BaseModel):
    sort_by: Optional[SortOptions] = None
    pagination_key: Optional[PaginationKey] = None
    page_size: int = DEFAULT_PAGE_SIZE


class GeneralListRequest(BaseModel):
    filter_by: Optional[FilterOptions] = None
    sort_by: Optional[SortOptions] = None
    pagination_key: Optional[PaginationKey] = None
    page_size: int = DEFAULT_PAGE_SIZE


class ListUserReviewingDatasetsRequest(BaseModel):
    pagination_key: Optional[PaginationKey] = None
    page_size: int = DEFAULT_PAGE_SIZE
    sort_by: Optional[SortOptions] = None
    filter_by: FilterOptions  # must provide reviewer user id.


class SchemaResponse(BaseModel):
    json_schema: Dict[str, Any]


class GenericFetchResponse(StatusResponse):
    item: Optional[Union[ItemBase, SeededItem]] = None
    roles: Optional[Roles] = None
    locked: Optional[bool] = None
    item_is_seed: Optional[bool] = None


class UntypedFetchResponse(StatusResponse):
    item: Optional[Dict[str, Any]] = None


class GenericListResponse(StatusResponse):
    # items
    items: Optional[List[Any]] = None
    seed_items: Optional[List[SeededItem]] = None
    unparsable_items: Optional[List[Dict[str, Any]]] = None

    # counts
    total_item_count: Optional[int] = None
    complete_item_count: Optional[int] = None
    seed_item_count: Optional[int] = None
    unparsable_item_count: Optional[int] = None
    not_authorised_count: Optional[int] = None

    # Pagination info
    pagination_key: Optional[PaginationKey] = None


class PaginatedListResponse(StatusResponse):
    # items
    items: Optional[List[Any]] = None
    # counts
    total_item_count: Optional[int] = None
    # Returned by previous paginations and used to indicate start of next query
    pagination_key: Optional[PaginationKey] = None


class PaginatedDatasetListResponse(StatusResponse):
    # items
    dataset_items: Optional[List[ItemDataset]] = None
    # counts
    total_dataset_count: Optional[int] = None
    # Returned by previous paginations and used to indicate start of next query
    pagination_key: Optional[PaginationKey] = None


class UpdateResponse(StatusResponse):
    # if this item has provenance enabled versioning AND the update was from
    # seed -> complete
    register_create_activity_session_id: Optional[str] = None


class GenericSeedResponse(StatusResponse):
    seeded_item: Optional[SeededItem] = None


class GenericCreateResponse(StatusResponse):
    created_item: Optional[ItemBase] = None

    # if this item has provenance enabled versioning, this job spins off from
    # creation to create and then lodge the CreateActivity
    register_create_activity_session_id: Optional[str] = None


class VersionResponse(BaseModel):
    new_version_id: str

    # Whenever versioning occurs, there is a job spun off to create the
    # associated Version activity, update various models, and lodge the
    # Version activity provenance
    version_job_session_id: str


class FetchItemRequest(BaseModel):
    item_id: IdentifiedResource


class ProxyItemFetchRequest(FetchItemRequest):
    username: str


class VersionRequest(BaseModel):
    # Which item to create version of
    id: str

    # Reason for the version being required
    reason: str


class UiSchemaResponse(StatusResponse):
    ui_schema: Optional[Dict[str, Any]] = None


class JsonSchemaResponse(StatusResponse):
    json_schema: Optional[Dict[str, Any]] = None


"""
========
ENTITIES
========
"""

# ==========================
# Dataset (Entity - Dataset)
# ==========================


class DatasetFetchResponse(GenericFetchResponse):
    item: Optional[Union[ItemDataset, SeededItem]]


class DatasetSeedResponse(GenericSeedResponse):
    None


class DatasetCreateResponse(GenericCreateResponse):
    created_item: Optional[ItemDataset]


class DatasetListResponse(GenericListResponse):
    items: Optional[List[ItemDataset]]

# =============================================
# Dataset Template (Entity - Dataset Template)
# =============================================


class DatasetTemplateFetchResponse(GenericFetchResponse):
    item: Optional[Union[ItemDatasetTemplate, SeededItem]]


class DatasetTemplateSeedResponse(GenericSeedResponse):
    None


class DatasetTemplateCreateResponse(GenericCreateResponse):
    created_item: Optional[ItemDatasetTemplate]


class DatasetTemplateListResponse(GenericListResponse):
    items: Optional[List[ItemDatasetTemplate]]

# ===============================================
# Software Base (Entity - Software) (NO ENDPOINT)
# ===============================================

# ===================================
# Model (Entity - Software <- Model)
# ===================================


class ModelFetchResponse(GenericFetchResponse):
    item: Optional[Union[ItemModel, SeededItem]]


class ModelSeedResponse(GenericSeedResponse):
    None


class ModelCreateResponse(GenericCreateResponse):
    created_item: Optional[ItemModel]


class ModelListResponse(GenericListResponse):
    items: Optional[List[ItemModel]]

# ============================================
# Registry Version (Activity - Registry Version)
# ============================================


class VersionFetchResponse(GenericFetchResponse):
    item: Optional[Union[ItemVersion, SeededItem]]


class VersionListResponse(GenericListResponse):
    items: Optional[List[ItemVersion]]

# ============================================
# Registry Create (Activity - Registry Create)
# ============================================


class CreateFetchResponse(GenericFetchResponse):
    item: Optional[Union[ItemCreate, SeededItem]]


class CreateListResponse(GenericListResponse):
    items: Optional[List[ItemCreate]]


# ===================================
# Study (Activity <- Study)
# ===================================

class StudyFetchResponse(GenericFetchResponse):
    item: Optional[Union[ItemStudy, SeededItem]]


class StudySeedResponse(GenericSeedResponse):
    None


class StudyCreateResponse(GenericCreateResponse):
    created_item: Optional[ItemStudy]


class StudyListResponse(GenericListResponse):
    items: Optional[List[ItemStudy]]


# ================================================================
# Workflow Template (Entity - Workflow Template) (NO ENDPOINT)
# ================================================================

# ======================================================================================
# Model Workflow Template (Entity - Workflow Template <- Model Workflow Template)
# ======================================================================================


class ModelRunWorkflowTemplateFetchResponse(GenericFetchResponse):
    item: Optional[Union[ItemModelRunWorkflowTemplate, SeededItem]]


class ModelRunWorkflowTemplateSeedResponse(GenericSeedResponse):
    None


class ModelRunWorkflowTemplateCreateResponse(GenericCreateResponse):
    created_item: Optional[ItemModelRunWorkflowTemplate]


class ModelRunWorkflowTemplateListResponse(GenericListResponse):
    items: Optional[List[ItemModelRunWorkflowTemplate]]


"""
==========
ACTIVITIES
==========
"""

# ====================================================
# Workflow Run (Activity - Workflow Run) (NO ENDPOINT)
# ====================================================

# =================================================
# Model Run (Activity - Workflow Run <- Model Run)
# =================================================


class ModelRunFetchResponse(GenericFetchResponse):
    item: Optional[Union[ItemModelRun, SeededItem]]


class ModelRunSeedResponse(GenericSeedResponse):
    None


class ModelRunCreateResponse(GenericCreateResponse):
    created_item: Optional[ItemModelRun]


class ModelRunListResponse(GenericListResponse):
    items: Optional[List[ItemModelRun]]


"""
========
AGENTS
========
"""

# ========================
# Person (Agent - Person)
# ========================


class PersonFetchResponse(GenericFetchResponse):
    item: Optional[Union[ItemPerson, SeededItem]]


class PersonSeedResponse(GenericSeedResponse):
    None


class PersonCreateResponse(GenericCreateResponse):
    created_item: Optional[ItemPerson]


class PersonListResponse(GenericListResponse):
    items: Optional[List[ItemPerson]]

# ====================================
# Organisation (Agent - Organisation)
# ====================================


class OrganisationFetchResponse(GenericFetchResponse):
    item: Optional[Union[ItemOrganisation, SeededItem]]


class OrganisationSeedResponse(GenericSeedResponse):
    None


class OrganisationCreateResponse(GenericCreateResponse):
    created_item: Optional[ItemOrganisation]


class OrganisationListResponse(GenericListResponse):
    items: Optional[List[ItemOrganisation]]


# Admin endpoints and interfaces

class BundledItem(BaseModel):
    id: str
    item_payload: Dict[str, Any]
    auth_payload: Dict[str, Any]
    lock_payload: Dict[str, Any]


class TableNames(BaseModel):
    resource_table_name: str
    auth_table_name: str
    lock_table_name: str


class RegistryExportResponse(StatusResponse):
    items: Optional[List[BundledItem]] = None


class RegistryImportSettings(BaseModel):
    import_mode: ImportMode

    # Should the items supplied be parsed before being
    # added/overwritten
    parse_items: bool = True

    # Always force explicit deletion flag even if mode enables it
    allow_entry_deletion: bool = False

    # Should the changes be written - true means no writing false means write
    # the changes
    trial_mode: bool = True


class RegistryRestoreRequest(RegistryImportSettings):
    table_names: TableNames


class RegistryImportRequest(RegistryImportSettings):
    items: List[BundledItem]


class RegistryImportStatistics(BaseModel):
    old_registry_size: int
    new_registry_size: int
    deleted_entries: int
    overwritten_entries: int
    new_entries: int


class RegistryImportResponse(StatusResponse):
    status: Status
    trial_mode: bool
    statistics: Optional[RegistryImportStatistics] = None
    failure_list: Optional[List[Tuple[str, Dict[str, Any]]]] = None


class ProvGraphRestoreRequest(BaseModel):
    trial_mode: bool = True
    abort_if_failures: bool = True
    items: List[BundledItem]


class ProvGraphRestoreResponse(StatusResponse):
    trial_mode: bool
    abort_if_failures: bool
    task_ids: Optional[List[str]] = None

# ==========================
# General Node Type
# ==========================


class Node(BaseModel):
    id: str
    item_category: ItemCategory
    item_subtype: ItemSubType



import collections
from typing import List

def duplicates(my_list: List[str]) -> List[str]:
    # Use collections counter object to get unique list of the inputted items 
    # and the fre of each to check for duplicates and return any :).
    return [h for h, count in collections.Counter(my_list).items() if count > 1]

def has_duplicates(my_list: List[str]) -> bool:
    return len(duplicates(my_list))>0


from ProvenaInterfaces.RegistryAPI import *
import pytest

# Ignore mypy errors as we are testing invalid models here and mypy
# doesn't know to look up and see "with self.assertRaises(ValueError):"
# The correct models will still fail if they aren't actually correct in the success case test
# mypy: ignore-errors


import pytest
from ProvenaInterfaces.RegistryAPI import SortOptions, SortType

def test_sort_options() -> None:
    # Test that begins_with can be used with valid sort types
    sort_options = SortOptions(sort_type=SortType.ACCESS_INFO_URI_BEGINS_WITH, begins_with="https://example.com")
    assert sort_options.sort_type == SortType.ACCESS_INFO_URI_BEGINS_WITH
    assert sort_options.begins_with == "https://example.com"

    # Test that begins_with cannot be used without specifying sort_type
    with pytest.raises(ValueError):
        SortOptions(begins_with="https://example.com")

    # Test that begins_with cannot be omitted when sort_type is ACCESS_INFO_URI_BEGINS_WITH
    with pytest.raises(ValueError):
        SortOptions(sort_type=SortType.ACCESS_INFO_URI_BEGINS_WITH)

    # Test that begins_with cannot be used with sort_type not in _valid_begins_with_sort_types
    with pytest.raises(ValueError):
        SortOptions(sort_type=SortType.CREATED_TIME, begins_with="https://example.com")


from pydantic import BaseModel
from enum import Enum
from typing import List


class ValueType(str, Enum):
    DESC = "DESC"
    URL = "URL"


class HandleProperty(BaseModel):
    type: ValueType
    value: str
    index: int


class Handle(BaseModel):
    id: str
    properties: List[HandleProperty]



from pydantic import model_validator, BaseModel
from typing_extensions import Self
from typing import Optional, Dict,  List

try:
    from ProvenaInterfaces.AsyncJobModels import *
except:
    from .AsyncJobModels import *

JOBS_USER_PREFIX = "/jobs/user"
JOBS_ADMIN_PREFIX = "/jobs/admin"

# API Actions and routes


PaginationKey = Dict[str, Any]


class JobsUserActions(str, Enum):
    FETCH = "FETCH"
    LIST = "LIST"
    LIST_BATCH = "LIST_BATCH"


JOBS_USER_ACTIONS_MAP: Dict[JobsUserActions, str] = {
    JobsUserActions.FETCH: "/fetch",
    JobsUserActions.LIST: "/list",
    JobsUserActions.LIST_BATCH: "/list_batch",
}


class JobsAdminActions(str, Enum):
    LAUNCH = "LAUNCH"
    FETCH = "FETCH"
    LIST = "LIST"
    LIST_BATCH = "LIST_BATCH"


JOBS_ADMIN_ACTIONS_MAP: Dict[JobsAdminActions, str] = {
    JobsAdminActions.LAUNCH: "/launch",
    JobsAdminActions.FETCH: "/fetch",
    JobsAdminActions.LIST: "/list",
    JobsAdminActions.LIST_BATCH: "/list_batch",
}

# ====
# USER
# ====

# -------
# get job
# -------

# req
# GET

# resp


class GetJobResponse(BaseModel):
    job: JobStatusTable

# ------------
# list by batch
# ------------


# req
# POST


class ListByBatchRequest(BaseModel):
    batch_id: str
    pagination_key: Optional[PaginationKey] = None
    limit: int = 10

# resp


class ListByBatchResponse(BaseModel):
    jobs: List[JobStatusTable]
    pagination_key: Optional[PaginationKey] = None

# ----------
# list jobs
# ----------


# req
# POST


class ListJobsRequest(BaseModel):
    pagination_key: Optional[PaginationKey] = None
    limit: int = 10

# resp


class ListJobsResponse(BaseModel):
    jobs: List[JobStatusTable]
    pagination_key: Optional[PaginationKey] = None

# ----------
# retry job
# ----------

# req
# POST

# resp


class RetryJobResponse(BaseModel):
    session_id: str


# =====
# ADMIN
# =====

# ----------
# launch job
# ----------

# req
# POST
class AdminLaunchJobRequest(BaseModel):
    # Who is this being lodged for? Defaults to token username
    username: Optional[str] = None
    # Does it need a batch id?
    request_batch_id: bool = False
    # Is it part of a batch?
    add_to_batch: Optional[str] = None
    # What type of job?
    job_type: JobType
    # What subtype?
    job_sub_type: JobSubType
    # Validated against job types expected payload
    job_payload: Dict[str, Any]

    @model_validator(mode='after')
    def check_batch_setup(self) -> Self:
        if self.request_batch_id and self.add_to_batch is not None:
            raise ValueError(
                "Cannot specify both a new batch and an existing batch id.")
        return self


# resp


class AdminLaunchJobResponse(BaseModel):
    # session id (unique)
    session_id: str

    # if requested, return the batch id
    batch_id: Optional[str] = None

# -------
# get job
# -------

# req
# GET (session_id)

# resp


class AdminGetJobResponse(BaseModel):
    job: JobStatusTable

# ----------
# list jobs
# ----------

# req
# POST


class AdminListJobsRequest(BaseModel):
    # Defaults to sorting by latest using global index

    # Optionally specify username which uses username index instead
    username_filter: Optional[str] = None
    limit: int = 10
    pagination_key: Optional[PaginationKey] = None

# resp


class AdminListJobsResponse(BaseModel):
    jobs: List[JobStatusTable]
    pagination_key: Optional[PaginationKey] = None


# -------------
# list by batch
# -------------


# req
# POST


class AdminListByBatchRequest(BaseModel):
    batch_id: str
    username_filter: Optional[str] = None
    pagination_key: Optional[PaginationKey] = None
    limit: int = 10

# resp


class AdminListByBatchResponse(BaseModel):
    jobs: List[JobStatusTable]
    pagination_key: Optional[PaginationKey] = None

# ----------
# retry job
# ----------

# req
# POST

# resp


class AdminRetryJobResponse(BaseModel):
    # session id (unique)
    session_id: str



import string
from typing import List, Optional, Tuple, Dict, Any
from pydantic import ConfigDict, BaseModel
from enum import Enum
# Interface exports prefer relative imports
# where as pip install prefers module level import
try:
    from ProvenaInterfaces.SharedTypes import Status, StatusResponse, ImportMode
    from ProvenaInterfaces.RegistryModels import Roles
except:
    from .SharedTypes import Status, StatusResponse, ImportMode
    from .RegistryModels import Roles

# The groups configuration includes references to the data store specific access
# for a given group
try:
    # from ProvenaInterfaces.DataStoreAPI import GroupActionsConfiguration
    None
except:
    None
    # from .DataStoreAPI import GroupActionsConfiguration


class IntendedUserType(str, Enum):
    """    IntendedUserType
        A list of possible types of users.
    """
    GENERAL = "GENERAL"
    ADMINISTRATOR = "ADMINISTRATOR"


class AccessLevel(str, Enum):
    """    AccessLevel
        Describes the access level of a role e.g. read/write/admin
    """
    READ = "READ"
    WRITE = "WRITE"
    ADMIN = "ADMIN"


class ComponentName(str, Enum):
    HANDLE_SERVICE = "handle-service"
    SYS_ADMIN = "sys-admin"
    ENTITY_REGISTRY = "entity-registry"
    JOB_SERVICE = "job-service"


class ComponentRole(BaseModel):
    """    ComponentRole
        Describes a role which includes all information 
        to understand the name, how to display/describe it 
        and what level it grants
    """
    # The proper keycloak role name
    role_name: str
    # The user friendly display name
    role_display_name: str
    # The level of access this is intended to grant
    # Please handle responsibly on the client side
    role_level: AccessLevel
    # Description of the purpose of the role
    description: str
    # Intended users
    intended_users: List[IntendedUserType]


class AuthorisationComponent(BaseModel):
    """    AuthorisationComponent
        An authorisation component is a component name
        along with a list of roles that component has.
    """
    # A component name, mainly used for displaying
    component_name: ComponentName
    # List of roles that component supports
    component_roles: List[ComponentRole]

    def get_role_at_level(self, desired_level: AccessLevel) -> ComponentRole:
        """    get_role_at_level
            For a given authorisation component, returns the access role which 
            grants the desired level of access. This can be used by dependent 
            components to work out what role name they should be looking for 
            given a particular level of access required.

            Arguments
            ----------
            desired_level : AccessLevel
                The access level this role should provide

            Returns
            -------
             : ComponentRole
                The component role which describes the matching role

            Raises
            ------
            Exception
                If no role is found with a matching level, then an exception is 
                thrown.

            See Also (optional)
            --------

            Examples (optional)
            --------
            read_only_role_name = DATA_STORE_COMPONENT.get_access_at_level(AccessLevel.READ).role_name
        """
        # Search for role with matching level
        for role in self.component_roles:
            if role.role_level == desired_level:
                # Return if found
                return role

        # Can't find the role - throw exception
        raise Exception(
            f'Failed to find a role that provides the correct access level\
              for component {self.component_name}.')


class AuthorisationModel(BaseModel):
    components: List[AuthorisationComponent]

    def get_role_level_for_component(self, component_name: ComponentName, role_level: AccessLevel) -> ComponentRole:
        """    get_role_level_for_component
            Given the component name and role level, will find the component role
            which provides that role level access.

            Arguments
            ----------
            component_name : str
                The component name e.g. data-store
            role_level : AccessLevel
                The desired level of access the role should grant

            Returns
            -------
             : ComponentRole
                The matching component role

            See Also (optional)
            --------

            Examples (optional)
            --------
        """
        # Search for matching component name
        # then use the get role at level component method
        for comp in self.components:
            if comp.component_name == component_name:
                return comp.get_role_at_level(role_level)

        raise Exception(
            'Unable to find a component with matching component name.')


JOB_COMPONENT = AuthorisationComponent(
    component_name=ComponentName.JOB_SERVICE,
    component_roles=[
        ComponentRole(
            role_name="job-service-admin",
            role_display_name="Job Service Admin",
            role_level=AccessLevel.ADMIN,
            description="Admin role which enables management of jobs including lodging.",
            intended_users=[
                IntendedUserType.ADMINISTRATOR
            ]
        ),
        ComponentRole(
            role_name="job-service-write",
            role_display_name="Job Service Write",
            role_level=AccessLevel.WRITE,
            description="Allows r/w and lodging of jobs",
            intended_users=[
                IntendedUserType.ADMINISTRATOR
            ]
        ),
        ComponentRole(
            role_name="job-service-read",
            role_display_name="Job Service Read",
            role_level=AccessLevel.READ,
            description="Allows reading of all user's jobs",
            intended_users=[
                IntendedUserType.ADMINISTRATOR
            ]
        )
    ]
)

HANDLE_SERVICE_COMPONENT = AuthorisationComponent(
    component_name=ComponentName.HANDLE_SERVICE,
    component_roles=[
        ComponentRole(
            role_name="handle-read",
            role_display_name="Handle service read only access",
            role_level=AccessLevel.READ,
            description="Allows access to read only functions of the handle service. This enables viewing, retrieving, listing of handles.",
            intended_users=[
                IntendedUserType.ADMINISTRATOR
            ]
        ),
        ComponentRole(
            role_name="handle-write",
            role_display_name="Handle service write access",
            role_level=AccessLevel.WRITE,
            description="Allows access to read and write functions of the handle service. This enables registering, viewing, retrieving, listing of handles.",
            intended_users=[
                IntendedUserType.ADMINISTRATOR
            ]
        ),
        ComponentRole(
            role_name="handle-admin",
            role_display_name="Handle service full admin access",
            role_level=AccessLevel.ADMIN,
            description="Allows full access to the handle service.",
            intended_users=[
                IntendedUserType.ADMINISTRATOR
            ]
        ),
    ]
)

SYS_ADMIN_COMPONENT = AuthorisationComponent(
    component_name=ComponentName.SYS_ADMIN,
    component_roles=[
        ComponentRole(
            role_name="sys-admin-read",
            role_display_name="Sys admin read only access",
            role_level=AccessLevel.READ,
            description="Allows read only access to system admin functions. This includes listing and viewing access requests.",
            intended_users=[
                IntendedUserType.ADMINISTRATOR
            ]
        ),
        ComponentRole(
            role_name="sys-admin-write",
            role_display_name="Sys admin write/read access",
            role_level=AccessLevel.WRITE,
            description="Allows read and write access to system admin functions. This includes listing, viewing, modifying and adding notes to access requests.",
            intended_users=[
                IntendedUserType.ADMINISTRATOR
            ]
        ),
        ComponentRole(
            role_name="sys-admin-admin",
            role_display_name="Sys admin full access",
            role_level=AccessLevel.ADMIN,
            description="Allows complete access to system admin functions. This includes listing, viewing, modifying, deleting and adding notes to access requests.",
            intended_users=[
                IntendedUserType.ADMINISTRATOR
            ]
        ),
    ]
)

ENTITY_REGISTRY_COMPONENT = AuthorisationComponent(
    component_name=ComponentName.ENTITY_REGISTRY,
    component_roles=[
        ComponentRole(
            role_name="entity-registry-read",
            role_display_name="Entity registry read only access",
            role_level=AccessLevel.READ,
            description="Allows read only access to the entity registry - list, view.",
            intended_users=[
                IntendedUserType.GENERAL,
                IntendedUserType.ADMINISTRATOR
            ]
        ),
        ComponentRole(
            role_name="entity-registry-write",
            role_display_name="Entity registry write/read access",
            role_level=AccessLevel.WRITE,
            description="Allows creation and modification of resources in the entity registry.",
            intended_users=[
                IntendedUserType.GENERAL,
                IntendedUserType.ADMINISTRATOR
            ]
        ),
        ComponentRole(
            role_name="entity-registry-admin",
            role_display_name="Entity registry full access",
            role_level=AccessLevel.ADMIN,
            description="Allows all actions against all entity registry resources.",
            intended_users=[
                IntendedUserType.ADMINISTRATOR
            ]
        ),
    ]
)

# Add more components as required here

AUTHORISATION_MODEL = AuthorisationModel(
    components=[
        HANDLE_SERVICE_COMPONENT,
        SYS_ADMIN_COMPONENT,
        ENTITY_REGISTRY_COMPONENT,
        JOB_COMPONENT
        # Add more here as required
    ]
)

# Start of response objects from auth API


class ReportComponentRole(ComponentRole):
    # inherits all component role fields
    # also includes whether this access is granted
    access_granted: bool


class ReportAuthorisationComponent(BaseModel):
    # A component name, mainly used for displaying
    component_name: ComponentName
    # List of roles that component supports
    component_roles: List[ReportComponentRole]


class AccessReport(BaseModel):
    components: List[ReportAuthorisationComponent]


class AccessReportResponse(BaseModel):
    status: Status
    report: AccessReport


class AccessRequestResponse(BaseModel):
    status: Status


class RequestStatus(str, Enum):
    # Initial state
    PENDING_APPROVAL = "PENDING_APPROVAL"

    # Approved and ready to be actioned
    APPROVED_PENDING_ACTION = "APPROVED_PENDING_ACTION"

    # Denied but not deleted - kept for auditing if desired
    DENIED_PENDING_DELETION = "DENIED_PENDING_DELETION"

    # Actioned and not deleted - kept for auditing if desired
    ACTIONED_PENDING_DELETION = "ACTIONED_PENDING_DELETION"


REQUEST_STATUS_TO_USER_EXPLANATION: Dict[RequestStatus, str] = {
    RequestStatus.PENDING_APPROVAL: "request received, awaiting action",
    RequestStatus.APPROVED_PENDING_ACTION: "request approved, action underway",
    RequestStatus.DENIED_PENDING_DELETION: "request denied",
    RequestStatus.ACTIONED_PENDING_DELETION: "request approved and completed",
}


class RequestAccessTableItem(BaseModel):
    # partition key
    username: str
    # sort key
    request_id: int
    # ttl attribute
    expiry: int
    # email address (== username when writing)
    email: str

    # other fields
    created_timestamp: int
    updated_timestamp: int
    status: RequestStatus
    ui_friendly_status: str

    # json serialised string of access report
    # for diff and for complete access report
    request_diff_contents: str
    complete_contents: str

    # comma separated list of notes
    notes: str


class AccessRequestList(BaseModel):
    items: List[RequestAccessTableItem]


class AccessRequestStatusChange(BaseModel):
    username: str
    request_id: int
    desired_state: RequestStatus
    additional_note: Optional[str] = None


class DeleteAccessRequest(BaseModel):
    username: str
    request_id: int


class RequestAddNote(BaseModel):
    note: str
    username: str
    request_id: int


class ChangeStateStatus(BaseModel):
    state_change: Status
    email_alert: Status


# Group management models

class GroupUser(BaseModel):
    # The username (required)
    username: str
    # the email (optional)
    email: Optional[str] = None
    # Optional first/last name
    first_name: Optional[str] = None
    last_name: Optional[str] = None


class UserGroupMetadata(BaseModel):
    # Unique ID to be used in dynamoDB table
    id: str
    # display name for the group e.g. Coconet Developers
    display_name: str
    # description of the groups purpose/users
    description: str

    # Optional additional metadata

    # Default access configuration for data store datasets - if a member of this
    # group creates a dataset, the dataset will have this group assigned with
    # the following default level. If this isn't provided, then no access will
    # be granted (other than the global default metadata read).
    default_data_store_access: Optional[Roles] = None


class UserGroup(UserGroupMetadata):
    # List of users
    users: List[GroupUser]


class ListGroupsResponse(StatusResponse):
    groups: Optional[List[UserGroupMetadata]] = None


class DescribeGroupResponse(StatusResponse):
    group: Optional[UserGroupMetadata] = None


class ListMembersResponse(StatusResponse):
    group: Optional[UserGroup] = None


class ListUserMembershipResponse(StatusResponse):
    groups: Optional[List[UserGroupMetadata]] = None


class CheckMembershipResponse(StatusResponse):
    is_member: Optional[bool] = None


class AddMemberResponse(StatusResponse):
    None


class RemoveMemberResponse(StatusResponse):
    None


class AddGroupResponse(StatusResponse):
    None


class RemoveGroupResponse(StatusResponse):
    None


class UpdateGroupResponse(StatusResponse):
    None


class RemoveMembersRequest(BaseModel):
    group_id: str
    member_usernames: List[str]


# import export revert

class GroupsImportStatistics(BaseModel):
    old_size: int
    new_size: int
    deleted_entries: int
    overwritten_entries: int
    new_entries: int


class GroupsExportResponse(StatusResponse):
    items: Optional[List[Dict[str, Any]]] = None


class GroupsImportSettings(BaseModel):
    import_mode: ImportMode

    # Should the items supplied be parsed before being
    # added/overwritten
    parse_items: bool = True

    # Always force explicit deletion flag even if mode enables it
    allow_entry_deletion: bool = False

    # Should the changes be written - true means no writing false means write
    # the changes
    trial_mode: bool = True


class GroupsRestoreRequest(GroupsImportSettings):
    None


class GroupsImportRequest(GroupsImportSettings):
    items: List[Dict[str, Any]]


class GroupsImportResponse(StatusResponse):
    trial_mode: bool
    statistics: Optional[GroupsImportStatistics] = None
    failure_list: Optional[List[Tuple[str, Dict[str, Any]]]] = None


class UsernamePersonLinkTableEntry(BaseModel):
    # both fields required

    # primary
    username: str
    # Standard-none, GSI-primary
    person_id: str
    model_config = ConfigDict(str_min_length=1)

# User -> Lookup
# Request = qstring no JSON body


class UserLinkUserLookupResponse(BaseModel):
    # Response

    # only included if present
    person_id: Optional[str] = None
    # included to specify if the link existed
    success: bool

# User -> Assign


class UserLinkUserAssignRequest(BaseModel):
    # username is implicit in token
    person_id: str
    model_config = ConfigDict(str_min_length=1)


class UserLinkUserValidateRequest(BaseModel):
    # username is implicit in token
    person_id: str
    model_config = ConfigDict(str_min_length=1)


class UserLinkUserValidateResponse(StatusResponse):
    # Just returns 200 OK but includes status for any issues
    pass


class UserLinkUserAssignResponse(BaseModel):
    # Just returns 200 OK, else 400/500 error with details
    pass

# Admin -> Lookup
# Request = qstring no JSON body


class AdminLinkUserLookupResponse(BaseModel):
    # Response

    # only included if present
    person_id: Optional[str] = None
    # included to specify if the link existed
    success: bool

# Admin -> Assign


class AdminLinkUserAssignRequest(BaseModel):
    # supply username
    username: str
    # username is implicit in token
    person_id: str
    # validate existence of item?
    validate_item: bool = True
    # force if existing
    force: bool = False
    model_config = ConfigDict(str_min_length=1)


class AdminLinkUserAssignResponse(BaseModel):
    # Just returns 200 OK, else 400/500 error with details
    pass


class AdminLinkUserClearResponse(BaseModel):
    # Just returns 200 OK, else 400/500 error with details
    pass


class UserLinkReverseLookupResponse(BaseModel):
    # the list of usernames that associate with that person_id, if any
    usernames: List[str]



from pydantic import field_validator, model_validator, BaseModel
from typing_extensions import Self
from typing import List, Optional, Dict, Any
from enum import Enum

# Base type alias for identified resources
IdentifiedResource = str

# Each of the various resource types which are associated with an ID
ModellerResource = IdentifiedResource
OrganisationResource = IdentifiedResource
ModelResource = IdentifiedResource
ModelRunWorkflowTemplateResource = IdentifiedResource
DatasetTemplateResource = IdentifiedResource
DataStoreDatasetResource = IdentifiedResource


class DatasetType(str, Enum):
    DATA_STORE = "DATA_STORE"


# Type alias placeholders for more complex structure if required later
ResourcePath = str
ResourceSpecification = ResourcePath


class TemplatedDataset(BaseModel):
    # What template does this dataset adhere to?
    dataset_template_id: DatasetTemplateResource

    # The actual dataset which must have type
    # matching the above
    dataset_id: DataStoreDatasetResource

    # Reference to which kind of type - currently hard coded as data store
    dataset_type: DatasetType

    # If the template specifies deferred resources, then they should be
    # specified here - the keys of this dictionary are the specified keys in the
    # deferred resource definition within the workflow template
    resources: Optional[Dict[str, ResourceSpecification]] = None

    def make_searchable(self) -> str:
        resource_items = []
        if self.resources:
            for k, v in self.resources.items():
                resource_items.append(k)
                resource_items.append(v)

        return " ".join(
            [
                self.dataset_template_id,
                self.dataset_id,
                " ".join(resource_items)
            ]
        )


class AssociationInfo(BaseModel):
    # What modeller produced this record - agent handle ID
    modeller_id: ModellerResource
    # Optionally, can specify an organisation that this model
    # run was requested by - agent ID in registry
    requesting_organisation_id: Optional[OrganisationResource] = None
    # TODO add non registrant attributions i.e. collaborators

    def make_searchable(self) -> str:
        elems = [self.modeller_id]
        if self.requesting_organisation_id:
            elems.append(self.requesting_organisation_id)
        return " ".join(elems)


class ModelRunRecord(BaseModel):
    # What workflow template does this model run fulfil?
    workflow_template_id: ModelRunWorkflowTemplateResource

    # RRAPIS-1580 - move this to the run itself to avoid weird churn on template
    # Optional to specify but first class - allows for whatever scheme makes
    # sense for user e.g. git hash
    model_version: Optional[str] = None

    # inputs (datasets, config, param files) - must fulfil template
    inputs: List[TemplatedDataset]
    outputs: List[TemplatedDataset]

    # custom user annotations - these are configured in the workflow template
    # and resolved here - can contain additional annotations, optional
    # annotations, but must have all required annotations
    annotations: Optional[Dict[str, str]] = None

    # pre canned annotations

    # model run display name
    display_name: str

    # provide a description of this run
    description: str

    # is this model run part of a study?
    study_id: Optional[str] = None

    # Associations of the record
    associations: AssociationInfo

    # Activity time metadata
    # unix timestamp (epoch)
    start_time: int
    # unix timestamp (epoch)
    end_time: int

    # validate start time is before end time using root validator
    @model_validator(mode='after')
    def validate_start_and_end(self) -> Self:
        start_time = self.start_time
        end_time = self.end_time

        if start_time is not None and end_time is not None:
            if start_time > end_time:
                raise ValueError(
                    f"The 'start_time' value ({start_time}) must be smaller than or equal to the 'end_time' value ({end_time}).")
        return self

    # validate display name and description are not empty strings using the same validator
    @field_validator("display_name", "description")
    @classmethod
    def string_must_not_be_empty(cls: Any, v: str) -> str:
        if v == "":
            raise ValueError("string must not be empty")
        return v

    def make_annotations_searchable(self) -> str:
        annotation_strings = []
        if self.annotations:
            for k, v in self.annotations.items():
                annotation_strings.append(k + " " + v)
        return " ".join(annotation_strings)

    @staticmethod
    def get_searchable_fields() -> List[str]:
        return [
            "workflow_template_id",
            "model_version",
            "display_name",
            "description",
            "inputs",
            "outputs",
            "annotations",
            "associations",
            "start_time",
            "end_time",
        ]

    def get_search_ready_object(self) -> Dict[str, str]:
        # helper function for search utilities
        fields: Dict[str, str] = {
            "workflow_template_id": self.workflow_template_id,
            "display_name": self.display_name,
            "model_version": self.model_version or "",
            "description": self.description or "",
            "inputs": " ".join([i.make_searchable() for i in self.inputs]),
            "outputs": " ".join([o.make_searchable() for o in self.outputs]),
            "annotations": self.make_annotations_searchable(),
            "associations": self.associations.make_searchable(),
            "start_time": str(self.start_time),
            "end_time": str(self.end_time),
        }
        return fields



from pydantic import BaseModel
from typing import List, Union

class Detail(BaseModel): 
        loc : List[Union[int, str]]
        msg : str 
        type : str
        
class HTTPValidationError(BaseModel):
        detail : List[Detail]


try:
    from ProvenaInterfaces.SharedTypes import StatusResponse
    from ProvenaInterfaces.ProvenanceModels import *
    from ProvenaInterfaces.RegistryAPI import ItemSubType
except:
    from .SharedTypes import StatusResponse
    from .ProvenanceModels import *

from typing import Optional, Any, Dict, Type
from pydantic import BaseModel, Field


class ProvenanceRecordInfo(BaseModel):
    # Handle ID
    id: str
    # Prov serialised document
    prov_json: str
    # Identified version of input format
    record: ModelRunRecord


class RegisterModelRunResponse(StatusResponse):
    session_id: Optional[str] = None


class SyncRegisterModelRunResponse(StatusResponse):
    record_info: ProvenanceRecordInfo


class RegisterBatchModelRunRequest(BaseModel):
    records: List[ModelRunRecord]


class RegisterBatchModelRunResponse(StatusResponse):
    session_id: Optional[str] = None


class LineageResponse(StatusResponse):
    record_count: Optional[int] = None
    graph: Optional[Dict[str, Any]] = None


class ConvertModelRunsResponse(StatusResponse):
    # list of model run records
    new_records: Optional[List[ModelRunRecord]] = None

    # list of job ids of validated existing jobs which don't need to be lodged
    existing_records: Optional[List[str]] = None

    warnings: Optional[List[str]] = None


class AddStudyLinkQueryParameters(BaseModel):
    # the model run that is being linked to the study
    model_run_id: str
    # the study that is being linked to the model run
    study_id: str


class AddStudyLinkResponse(StatusResponse):
    # the model run that was linked to the study
    model_run_id: str
    # the study that was linked to the model run
    study_id: str
    # the session id for the provenance graph update job
    session_id: str


class PostUpdateModelRunInput(BaseModel):
    model_run_id: str
    reason: str
    record: ModelRunRecord


class PostUpdateModelRunResponse(BaseModel):
    session_id: str

class GenerateReportRequest(BaseModel):
    id: str
    item_subtype: ItemSubType
    # Greater than or equal to 1 and less than equal to 3
    depth: int = Field(ge=1, le=3)

class PostDeleteGraphRequest(BaseModel):
    record_id: str
    trial_mode: bool = False

class PostDeleteGraphResponse(BaseModel):
    # Diff list - each entry is a diff action with possible other metadata
    diff: List[Dict[str, Any]]



from pydantic import BaseModel
from typing import Optional
try:
    from ProvenaInterfaces.SharedTypes import StatusResponse
    from ProvenaInterfaces.HandleModels import *
except:
    from .SharedTypes import StatusResponse
    from .HandleModels import *


class ValueRequestBase(BaseModel):
    value_type: ValueType
    value: str


class AddValueRequest(ValueRequestBase):
    id: str


class AddValueIndexRequest(ValueRequestBase):
    id: str
    index: int


class MintRequest(ValueRequestBase):
    pass


class ModifyRequest(BaseModel):
    id: str
    index: int
    value: str


class RemoveRequest(BaseModel):
    id: str
    index: int


class ListResponse(BaseModel):
    ids: List[str]


MintResponse = Handle
AddValueResponse = Handle
AddValueIndexResponse = Handle
GetResponse = Handle
ModifyResponse = Handle
RemoveResponse = Handle



from pydantic import BaseModel
from enum import Enum
from typing import Optional, List


class UserInfo(BaseModel):
    username: str
    email: str
    roles: List[str]


# Encrypted user info is just a string - but indicates it will decrypt into
# UserInfo
EncryptedUserInfo = str


class Status(BaseModel):
    success: bool
    details: str


class StatusResponse(BaseModel):
    status: Status


class ImportMode(str, Enum):
    # Will never change existing entries Can only add new entries
    ADD_ONLY = "ADD_ONLY"

    # Will add new entries, or overwrite existing if they match
    ADD_OR_OVERWRITE = "ADD_OR_OVERWRITE"

    # Only allows updating of existing entries
    OVERWRITE_ONLY = "OVERWRITE_ONLY"

    # Will add new entries, overwrite existing and enforce there being no
    # unmatched entries
    SYNC_ADD_OR_OVERWRITE = "SYNC_ADD_OR_OVERWRITE"

    # Will perform a full sync meaning new additions added, existing overwritten
    # and left over unmatched entries deleted
    SYNC_DELETION_ALLOWED = "SYNC_DELETION_ALLOWED"


class VersionDetails(BaseModel):
    commit_id: Optional[str] = None
    commit_url: Optional[str] = None
    tag_name: Optional[str] = None
    release_title: Optional[str] = None
    release_url: Optional[str] = None



from pydantic import BaseModel
from typing import Optional, List, Any, Dict
from enum import Enum

try:
    from ProvenaInterfaces.SharedTypes import StatusResponse
except:
    from .SharedTypes import StatusResponse


class SearchResultType(str, Enum):
    DATASET = "DATASET"
    REGISTRY_ITEM = "REGISTRY_ITEM"


class QueryResult(BaseModel):
    id: str
    score: float


class MixedQueryResult(QueryResult):
    type: SearchResultType


class QueryResults(StatusResponse):
    results: Optional[List[QueryResult]] = None
    warnings: Optional[List[str]] = None


class MixedQueryResults(StatusResponse):
    results: Optional[List[MixedQueryResult]] = None
    warnings: Optional[List[str]] = None



from pydantic import field_validator, model_validator, ConfigDict, BaseModel, AnyHttpUrl, EmailStr, validator, Field, schema, AnyUrl
from typing_extensions import Self
from enum import Enum
from typing import Optional, List, Tuple, TypeVar, Dict, Type, Any, Set, Generic, Type
import itertools
from isodate import parse_duration  # type: ignore
from datetime import date

try:
    from ProvenaInterfaces.helpers.helpers import duplicates
except:
    from .helpers.helpers import duplicates

# Type alias to denote that a string references an ID
IdentifiedResource = str


class ItemCategory(str, Enum):
    ACTIVITY = "ACTIVITY"
    AGENT = "AGENT"
    ENTITY = "ENTITY"


class ItemSubType(str, Enum):
    # Activities
    WORKFLOW_RUN = "WORKFLOW_RUN"
    MODEL_RUN = "MODEL_RUN"  # is a WORKFLOW_RUN
    STUDY = "STUDY"
    CREATE = "CREATE"
    VERSION = "VERSION"

    # Agents
    PERSON = "PERSON"  # ✓
    ORGANISATION = "ORGANISATION"  # ✓

    # Entities
    SOFTWARE = "SOFTWARE"  # ✓
    MODEL = "MODEL"  # is a SOFTWARE # ✓
    WORKFLOW_TEMPLATE = "WORKFLOW_TEMPLATE"  # ✓
    MODEL_RUN_WORKFLOW_TEMPLATE = \
        "MODEL_RUN_WORKFLOW_TEMPLATE"  # is a WORKFLOW_DEF'N # ✓
    DATASET = "DATASET"  # ✓
    DATASET_TEMPLATE = "DATASET_TEMPLATE"  # ✓


"""
===========
LOCK MODELS
===========
"""


class LockActionType(str, Enum):
    LOCK = "LOCK"
    UNLOCK = "UNLOCK"


class LockEvent(BaseModel):
    # Locked or unlocked
    action_type: LockActionType
    # Who did it?

    # user
    username: str
    # email
    email: Optional[str] = None

    # Why did they lock/unlock it
    reason: str
    # When did this happen? (unix timestamp)
    timestamp: int


class LockInformation(BaseModel):
    # is the resource locked?
    locked: bool
    # what is the history of lock actions
    history: List[LockEvent]

# maps to the lock table deployed in parallel to main registry table


class LockTableEntry(BaseModel):
    # the id (handle) primary key
    id: str
    # the access settings
    lock_information: LockInformation


"""
============
AUTH MODELS 
============
"""

# A role is a string
Role = str

# A role list is a collection of roles
Roles = List[Role]

# Basic read/write for metadata - common to all entities
METADATA_READ_ROLE = "metadata-read"
METADATA_WRITE_ROLE = "metadata-write"

metadata_read_write: Roles = [
    METADATA_READ_ROLE,
    METADATA_WRITE_ROLE
]

# Roles for datasets
DATASET_READ_ROLE = "dataset-data-read"
DATASET_WRITE_ROLE = "dataset-data-write"

dataset_data_read_write: Roles = [
    DATASET_READ_ROLE,
    DATASET_WRITE_ROLE,
]

# owner role - this is a special role
ADMIN_ROLE = "admin"
admin_role: Roles = [
    ADMIN_ROLE
]


# These are role lists for the various entity types
ENTITY_BASE_ROLE_LIST = metadata_read_write + admin_role
DATASET_ROLE_LIST = ENTITY_BASE_ROLE_LIST + dataset_data_read_write


ALL_ROLES = set(DATASET_ROLE_LIST + ENTITY_BASE_ROLE_LIST)


class DescribedRole(BaseModel):
    role_display_name: str
    role_name: str
    description: str
    also_grants: List[str]


# for each role - provides some metadata
ROLE_METADATA_MAP: Dict[str, DescribedRole] = {
    METADATA_READ_ROLE: DescribedRole(
        role_display_name="Metadata Read",
        role_name=METADATA_READ_ROLE,
        also_grants=[],
        description="Enables visibility of a resource and it's metadata."
    ),
    METADATA_WRITE_ROLE: DescribedRole(
        role_display_name="Metadata Write",
        role_name=METADATA_WRITE_ROLE,
        also_grants=[METADATA_READ_ROLE],
        description="Enables the modification of a resource's metadata."
    ),
    DATASET_READ_ROLE: DescribedRole(
        role_display_name="Dataset Data Read",
        role_name=DATASET_READ_ROLE,
        also_grants=[METADATA_READ_ROLE],
        description="Enables the downloading of the contents of a dataset."
    ),
    DATASET_WRITE_ROLE: DescribedRole(
        role_display_name="Dataset Data Write",
        role_name=DATASET_WRITE_ROLE,
        also_grants=[METADATA_READ_ROLE,
                     METADATA_WRITE_ROLE, DATASET_READ_ROLE],
        description="Enables the uploading, modification and deletion of dataset files."
    ),
    ADMIN_ROLE: DescribedRole(
        role_display_name="Admin",
        role_name=ADMIN_ROLE,
        also_grants=[METADATA_READ_ROLE, METADATA_WRITE_ROLE,
                     DATASET_READ_ROLE, DATASET_WRITE_ROLE],
        description="Enables all actions on a resource and it's contents, including changes to access settings."
    )
}

# Check that all roles are described
for role in ALL_ROLES:
    assert role in ROLE_METADATA_MAP


class AccessSettings(BaseModel):
    # who owns this thing - established via username (persistent unchangeable unique keycloak user identifier)
    owner: str

    # What can all registry users do bounded by registry access
    general: Roles

    # map of Group ID from Auth API -> AccessModel
    groups: Dict[str, Roles]


# maps to the auth table deployed in parallel to main registry table
class AuthTableEntry(BaseModel):
    # the id (handle) primary key
    id: str
    # the access settings
    access_settings: AccessSettings


class DomainInfoBase(BaseModel):
    # This is the base for the seed item
    # input - each sub type will extend this
    # to specify additional required domain
    # information

    # what is the minimal display name
    # for this item
    display_name: str

    # all item types can also contain customised user metadata (optionally)
    # for datasets - this is kept in sync with the duplicated field in the collection format
    user_metadata: Optional[Dict[str, str]] = Field(
        None,
        title="Custom User Metadata",
        description="Optionally provide a collection of key value annotations describing this resource."
    )

    @staticmethod
    def get_searchable_fields() -> List[str]:
        return ['display_name', 'user_metadata']

    def get_search_ready_object(self) -> Dict[str, str]:
        base = {
            'display_name': self.display_name,
        }
        if self.user_metadata is not None:
            compacted = ""
            for k, v in self.user_metadata.items():
                compacted += f"{k} {v} "
            base['user_metadata'] = compacted
        else:
            base['user_metadata'] = ""

        return base


class RecordType(str, Enum):
    SEED_ITEM = "SEED_ITEM"
    COMPLETE_ITEM = "COMPLETE_ITEM"


DomainInfoTypeVar = TypeVar('DomainInfoTypeVar', bound=DomainInfoBase)


class HistoryEntry(BaseModel, Generic[DomainInfoTypeVar]):
    # unique id within scope of record - incremented
    id: int
    # When was this entry created
    timestamp: int
    # Why was this update performed?
    reason: str
    # What is the username of the user who performed this update?
    username: str
    # The contents of the domain info as of this update
    item: DomainInfoTypeVar


def unique_history_ids(history_list: List[HistoryEntry[Any]]) -> None:
    ids = list(map(lambda h: h.id, history_list))
    assert len(ids) == len(set(ids)), "Non unique ID list in item history"


class HistoryBase(BaseModel, Generic[DomainInfoTypeVar]):
    # History is just a list of history entries - this is a generic type base class
    history: List[HistoryEntry[DomainInfoTypeVar]]


class VersioningInfo(BaseModel):
    previous_version: Optional[str] = None
    version: int
    reason: Optional[str] = None
    next_version: Optional[str] = None

    @model_validator(mode='after')
    def check_reason_provided(self) -> Self:
        if self.previous_version is not None:
            if self.reason is None:
                raise ValueError(
                    "Items with a previous version must specify a reason for the updated version.")
        return self


class WorkflowLinks(BaseModel):
    # Create prov session id spinoff (optionally)
    create_activity_workflow_id: Optional[str] = None

    # Version prov session id spinoff
    version_activity_workflow_id: Optional[str] = None


class RecordInfo(BaseModel):
    # Handle ID
    # optional for unseeded identities
    id: str

    # owners username - this links to a registered Person through the link service
    owner_username: str

    # when was this item created
    # unix timestamp
    # optional - added in at seed time
    created_timestamp: int

    # when was this item last updated
    # optional - added in on update/seed
    updated_timestamp: int

    # what is the overall type
    item_category: ItemCategory

    # what is the sub type
    item_subtype: ItemSubType

    # is this seed or complete
    record_type: RecordType

    # Not all types have workflow links - only prov versioning enabled
    workflow_links: Optional[WorkflowLinks] = None

    # Versioning (all optional)
    versioning_info: Optional[VersioningInfo] = None

    @staticmethod
    def get_searchable_fields() -> List[str]:
        # this is a special elastic search syntax
        return ['id^4', 'item_category', 'item_subtype']

    def get_search_ready_object(self) -> Dict[str, str]:
        return {
            'id': self.id,
            'item_category': self.item_category.value,
            'item_subtype': self.item_subtype.value,
            'owner_username': self.owner_username
        }


class SeededItem(RecordInfo):
    # inherits all values from RecordInfo including search term generators
    None


class ItemBase(RecordInfo, DomainInfoBase, HistoryBase[DomainInfoBase]):
    def get_search_ready_object(self) -> Dict[str, str]:
        record_info = RecordInfo.parse_obj(self.dict())
        ri_base = record_info.get_search_ready_object()
        domain_info = DomainInfoBase.parse_obj(self.dict())
        di_base = domain_info.get_search_ready_object()
        ri_base.update(di_base)
        return ri_base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        # return combination of record info and domain info
        return RecordInfo.get_searchable_fields() +\
            DomainInfoBase.get_searchable_fields()


# Entity/activity/agent extension of item base


class EntityBase(ItemBase):
    # override item category
    item_category: ItemCategory = ItemCategory.ENTITY


class AgentBase(ItemBase):
    # override item category
    item_category: ItemCategory = ItemCategory.AGENT


class ActivityBase(ItemBase):
    # override item category
    item_category: ItemCategory = ItemCategory.ACTIVITY


"""
========
ENTITIES
========
"""

# =============================================
# Dataset Template (Entity - Dataset Template)
# =============================================


class ResourceUsageType(str, Enum):
    # This is a parameter file
    PARAMETER_FILE = "PARAMETER_FILE"
    # This is a config file
    CONFIG_FILE = "CONFIG_FILE"
    # This is forcing data
    FORCING_DATA = "FORCING_DATA"
    # This is unspecified general input data
    GENERAL_DATA = "GENERAL_DATA"


class ResourceSpatialMetadata(BaseModel):
    # TODO how do we define spatial information?
    None


class ResourceTemporalMetadata(BaseModel):
    # TODO how do we define temporal information?

    # temporal extent
    # unix timestamp
    start_time: int
    # unix timestamp
    end_time: int

    # temporal jump
    # timestep
    # TODO collaborate on whether this is useful
    # interval: int


class DatasetParameter(BaseModel):
    # TODO collaborate on this - starting ideas
    parameter_name: Optional[str] = None
    column_name: Optional[str] = None
    column_index: Optional[int] = None
    description: Optional[str] = None
    vocabulary_id: Optional[str] = None


# class BaseResourceInformation(BaseModel):
#    # e.g. "This file/folder describes the expected coral values under xyz
#    # assumptions"
#    description: str
#
#    # Will the model run be rejected if no path is provided for this resource?
#    optional: Optional[bool] = None
#
#    # usage - what is this thing being used for
#    usage_type: ResourceUsageType
#
#    # type of thing - assumes file but could be folder
#    is_folder: Optional[bool] = None
#
#    # additional metadata (free form KVPs for users to add things)
#    additional_metadata: Optional[Dict[str, str]] = None
#
#    # Simplifying payload for now...
#
#    # e.g. ".csv"
#    # extension: Optional[str] = None
#
#    # description of parameters that this dataset reports
#    # parameters: Optional[List[DatasetParameter]] = None
#
#    # associated temporal/spatial information
#    # spatial_metadata: Optional[ResourceSpatialMetadata] = None
#    # temporal_metadata: Optional[ResourceTemporalMetadata] = None
#
#    # size estimates (MB)
#    # size_estimate: Optional[int] = None
#    # size_min: Optional[int] = None
#    # size_max: Optional[int] = None
#
#    # enforce size constraints?
#    # enforce_size_constraints: Optional[bool] = False


class DefinedResource(BaseModel):
    # Contains all base information + the known path relative to the dataset
    # base
    path: str

    # e.g. "This file/folder describes the expected coral values under xyz
    # assumptions"
    description: str

    # usage - what is this thing being used for
    usage_type: ResourceUsageType

    # Will the model run be rejected if no path is provided for this resource?
    optional: Optional[bool] = None

    # type of thing - assumes file but could be folder
    is_folder: Optional[bool] = None

    # additional metadata (free form KVPs for users to add things)
    additional_metadata: Optional[Dict[str, str]] = None

    def get_search_ready_string(self) -> str:
        return " ".join(
            [
                # resource path
                self.path,
                # description
                self.description,
                # annotations
                " ".join([f"{key} {value}" for key, value in (
                    self.additional_metadata.items() if self.additional_metadata else {})])
            ]
        )


class DeferredResource(BaseModel):
    # Contains all the base information and,

    # a resource key - at model run time, this resource key must be matched with
    # a path that fulfils it
    key: str

    # e.g. "This file/folder describes the expected coral values under xyz
    # assumptions"
    description: str

    # usage - what is this thing being used for
    usage_type: ResourceUsageType

    # Will the model run be rejected if no path is provided for this resource?
    optional: Optional[bool] = None

    # type of thing - assumes file but could be folder
    is_folder: Optional[bool] = None

    # additional metadata (free form KVPs for users to add things)
    additional_metadata: Optional[Dict[str, str]] = None

    def get_search_ready_string(self) -> str:
        return " ".join(
            [
                # resource key
                self.key,
                # description
                self.description,
                # annotations
                " ".join([f"{key} {value}" for key, value in (
                    self.additional_metadata.items() if self.additional_metadata else {})])
            ]
        )


class DatasetTemplateDomainInfo(DomainInfoBase):
    # Provide a basic description of the nature of this template - if the
    # resources are adequately described within the resource payloads then this
    # is optional
    description: Optional[str] = None

    # These are resources known/defined at template time - covers the use case
    # in which the user has a pre-specified file path/name within the dataset
    # which is known at template time consistently
    defined_resources: List[DefinedResource] = []

    # Deferred resources represents a 'placeholder' in which the exact file path
    # information will be supplied at model run time.
    deferred_resources: List[DeferredResource] = []

    @model_validator(mode='after')
    def check_unique_deferred_keys(self) -> Self:
        if self.deferred_resources is not None:
            key_set: Set[str] = set()
            for resource in self.deferred_resources:
                key_set.add(resource.key)
            if len(key_set) != len(self.deferred_resources):
                raise ValueError(
                    "Non unique resource_key properties in provided deferred resources.")
        return self


class ItemDatasetTemplate(EntityBase, DatasetTemplateDomainInfo):
    item_subtype: ItemSubType = ItemSubType.DATASET_TEMPLATE

    # override history
    history: List[HistoryEntry[DatasetTemplateDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
            "description": self.description or "",
            "deferred_resources": " ".join([
                r.get_search_ready_string() for r in self.deferred_resources
            ]),
            "defined_resources": " ".join([
                r.get_search_ready_string() for r in self.defined_resources
            ]),
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        return ItemBase.get_searchable_fields() + [
            "description",
            "deferred_resources",
            "defined_resources",
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v


# ==================================
# Software Base (Entity - Software)
# ==================================
class SoftwareDomainInfo(DomainInfoBase):
    # name of piece of software
    name: str

    # description
    # what is this software?
    description: str

    # URL to more information
    documentation_url: AnyHttpUrl

    # where to find the source
    source_url: AnyHttpUrl

    # TODO this can be expanded as per https://confluence.csiro.au/display/RRAPIS/Provenance+Entities
    # keep this generic information about software rather than model specific


class ItemSoftware(EntityBase, SoftwareDomainInfo):
    # override sub type
    item_subtype: ItemSubType = ItemSubType.SOFTWARE

    # override history
    history: List[HistoryEntry[SoftwareDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
            "name": self.name,
            "description": self.description,
            "documentation_url": str(self.documentation_url),
            "source_url": str(self.source_url),
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        return ItemBase.get_searchable_fields() + [
            "name",
            "description",
            "documentation_url",
            "source_url"
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v


# ===================================
# Model (Entity - Software <- Model)
# ===================================


class ModelDomainInfo(SoftwareDomainInfo):
    # TODO specialty model information here
    # maybe reference to registered model?
    # e.g. coconet?
    None


class ItemModel(EntityBase, ModelDomainInfo):
    # override sub type
    item_subtype: ItemSubType = ItemSubType.MODEL

    # override history
    history: List[HistoryEntry[ModelDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
            "name": self.name,
            "description": self.description,
            "documentation_url": str(self.documentation_url),
            "source_url": str(self.source_url),
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        return ItemBase.get_searchable_fields() + [
            "name",
            "description",
            "documentation_url",
            "source_url"
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v

# ===================================================
# Workflow Template (Entity - Workflow Template)
# ===================================================


class AutomationSchedule(BaseModel):
    # TODO define this
    None


class WorkflowTemplateAnnotations(BaseModel):
    required: List[str] = []
    optional: List[str] = []

    # Check that these keys are unique
    @model_validator(mode='after')
    def check_unique_keys(self) -> Self:
        if self.required is not None and self.optional is not None:
            key_set: Set[str] = set(self.required).union(set(self.optional))
            if not (len(key_set) == len(self.required) + len(self.optional)):
                raise ValueError(
                    "Provided required and/or optional field keys are non-unique.")
        return self

    def make_searchable(self) -> str:
        return " ".join(self.required) + " ".join(self.optional)


class TemplateResource(BaseModel):
    template_id: IdentifiedResource
    optional: Optional[bool] = None


class WorkflowTemplateDomainInfo(DomainInfoBase):
    # Specify the software ID and version at time of template the validity of a
    # definition is tied to the lifecycle of the software
    software_id: IdentifiedResource

    # RRAPIS-1580 - move this to the run itself to avoid weird churn on template
    # software_version: str

    # inputs? List of dataset templates
    input_templates: List[TemplateResource] = []

    # outputs? List of dataset templates
    output_templates: List[TemplateResource] = []

    # Expected KVPs
    annotations: Optional[WorkflowTemplateAnnotations] = None

    @field_validator('input_templates', 'output_templates')
    @classmethod
    def template_ids_unique_per_end(cls, v: List[TemplateResource]) -> List[TemplateResource]:
        id_list = [template.template_id for template in v]
        if len(duplicates(id_list)):
            raise ValueError(
                f"Cannot support duplicate templates being used twice in just input templates, or twice in just output templates. Duplicates: {duplicates(id_list)}.")
        return v


class ItemWorkflowTemplate(EntityBase, WorkflowTemplateDomainInfo):
    # override sub type
    item_subtype: ItemSubType = ItemSubType.WORKFLOW_TEMPLATE

    # override history
    history: List[HistoryEntry[WorkflowTemplateDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
            "software_id": self.software_id,
            # RRAPIS-1580 moves this to model run payload
            # "software_version": self.software_version,
            "input_templates": " ".join([t.template_id for t in self.input_templates]),
            "output_templates": " ".join([t.template_id for t in self.output_templates]),
            "annotations": self.annotations.make_searchable() if self.annotations else ""
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        return ItemBase.get_searchable_fields() + [
            "software_id",
            # RRAPIS-1580 moves this to model run payload
            # "software_version",
            "input_templates",
            "output_templates",
            "annotations"
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v

# ======================================================================================
# Dataset (Entity <- Dataset)
# ======================================================================================


class CollectionFormatAssociations(BaseModel):
    """
    Please provide information about the people and organisations related to this dataset.
    """
    # Registered organisation id
    organisation_id: str = Field(
        ...,
        title="Record Creator Organisation",
        description="Please select the organisation on behalf of which you are registering this data."
    )

    # ID of the registered Person Custodian
    data_custodian_id: Optional[str] = Field(
        None,
        title="Dataset Custodian",
        description="Please select the Registered Person who is the dataset custodian.",
    )

    # Free text info for point of contact
    point_of_contact: Optional[str] = Field(
        None,
        title="Point of Contact",
        description="Please provide a point of contact for enquiries about this data e.g. email address. Please ensure you have sought consent to include these details in the record.",
    )
    model_config = ConfigDict(
        str_min_length=1, extra="forbid", title="Dataset Associations")


class AccessInfo(BaseModel):
    """
    Please specify whether data is going to be stored in the Data Store, or referenced from an existing, externally hosted, source.
    """
    # is this going to be stored in the provena data store?
    reposited: bool = Field(
        True,
        title="Data Storage",
        description="Is the data going to be stored in the Data Store?"
    )

    # What is the best URI to access this data
    uri: Optional[AnyUrl] = Field(
        default=None,
        title="Dataset URI",
        description="Provide the best URI of the externally hosted data.",
        json_schema_extra={"format": "uri"}
    )

    # Instructions/advice on how to retrieve it
    description: Optional[str] = Field(
        None,
        title="Access Description",
        description="Provide information about how to access the externally hosted data, or other relevant notes."
    )

    @model_validator(mode='after')
    def check_external_access_provided(self) -> Self:
        # This validator ensures that if reposited is False, external access
        # info is not None and fields are populated

        # arbitrary change
        assert self.reposited is not None

        if not self.reposited:
            if self.uri is None:
                raise ValueError(
                    "Must provide a URI for external access if data is not reposited in the Data Store.")
            if self.description is None:
                raise ValueError(
                    "Must provide a description for external access if data is not reposited in the Data Store.")
        else:
            # it is reposited, ensure URI and description are None
            if self.uri is not None:
                raise ValueError(
                    "Cannot provide a URI for external access if data is reposited in the Data Store.")
            if self.description is not None:
                raise ValueError(
                    "Cannot provide a description for external access if data is reposited in the Data Store.")
        return self

    model_config = ConfigDict(
        str_min_length=1, extra="forbid", title="Access Info")


class OptionallyRequiredCheck(BaseModel):
    # default to False to ensure forms require explicit inclusion
    relevant: bool = False
    # default to true so forms default to satisfied state
    obtained: bool = False


# both of these controls are instances of optionally required check
class ExportControls(OptionallyRequiredCheck):
    """
    Is this dataset subject to any export controls permits? If so, has this dataset cleared any required due diligence checks and have you obtained any required permits?
    """
    model_config = ConfigDict(extra="forbid", title="Export Controls")


class IndigenousKnowledgeCheck(OptionallyRequiredCheck):
    """
    Does this dataset contain Indigenous Knowledge? If so, do you have consent from the relevant Aboriginal and Torres Strait Islander communities for its use and access via this data store?
    """
    model_config = ConfigDict(
        extra="forbid", title="Indigenous Knowledge and Consent")


class DatasetEthicsRegistrationCheck(OptionallyRequiredCheck):
    """
    Does this dataset include any human data or require ethics/privacy approval for its registration? If so, have you included any required ethics approvals, consent from the participants and/or appropriate permissions to register this dataset in this information system?
    """
    model_config = ConfigDict(
        extra="forbid", title="Dataset Registration Ethics and Privacy")


class DatasetEthicsAccessCheck(OptionallyRequiredCheck):
    """
    Does this dataset include any human data or require ethics/privacy approval for enabling its access by users of the information system? If so, have you included any required consent from the participants and/or appropriate permissions to facilitate access to this dataset in this information system?
    """
    model_config = ConfigDict(
        str_min_length=1, extra="forbid", title="Dataset Access Ethics and Privacy")


class CollectionFormatSpatialInfo(BaseModel):
    """
    If your dataset includes spatial data, you can indicate the coverage, resolution and extent of this spatial data.
    """
    # What is the spatial coverage of the dataset? EWKT TODO validator
    coverage: Optional[str] = Field(
        None,
        title="Spatial Coverage",
        # TODO clarity here re: format and expectations of geometry features e.g. Geometry Collection?
        description="The geographic area applicable to the data asset. Please specify spatial coverage using the EWKT format.",
        # 50000 characters limit TODO justify this limit
        max_length=50000
    )

    # Decimal Degrees spatial resolution TODO validator
    resolution: Optional[str] = Field(
        None,
        title="Spatial Resolution",
        description="The spatial resolution applicable to the data asset. Please use the Decimal Degrees standard."
    )

    # Spatial extent i.e. bounding box - EWKT TODO validator
    extent: Optional[str] = Field(
        None,
        title="Spatial Extent",
        # TODO clarity here re: format and expectations of geometry features e.g. Geometry Collection, Polygon?
        description="The range of spatial coordinates applicable to the data asset. Please provide a bounding box extent using the EWKT format.",
        # 50000 characters limit TODO justify this limit
        max_length=50000
    )

    @field_validator('resolution')
    @classmethod
    def validate_spatial_resolution(cls: Any, v: Optional[str]) -> Optional[str]:
        # None is okay as the field is optional
        if v is None:
            return v

        # val must be parsable as float
        try:
            val = float(v)
            assert val > 0, f"Spatial resolution must be non-negative and greater than than 0. Got '{val}'."
        except Exception as e:
            raise ValueError(
                f"Invalid spatial resolution. The value must conform to the Decimal Degrees format. Please provide a positive decimal value. Failed to parse the spatial resolution as a positive float. Details: {e}")
        return v
    model_config = ConfigDict(
        str_min_length=1, extra="forbid", title="Dataset Spatial Information")


class TemporalDurationInfo(BaseModel):
    """
    You can specify the duration of temporal coverage by including both start and end date.
    """
    begin_date: date = Field(
        ...,
        title="Start Date",
        description="Please provide the start date of the dataset's temporal coverage."
    )
    end_date: date = Field(
        ...,
        title="End Date",
        description="Please provide the end date of the dataset's temporal coverage."
    )

    # validator to ensure that if both are set, they are in the correct order
    @model_validator(mode='after')
    def check_dates_in_order(self) -> Self:
        if self.begin_date is not None and self.end_date is not None:
            if self.begin_date > self.end_date:
                raise ValueError(
                    "Begin Date must be before or equal to End Date.")
        return self

    model_config = ConfigDict(
        str_min_length=1, extra="forbid", title="Temporal Duration")


class CollectionFormatTemporalInfo(BaseModel):
    """
    If your dataset provides data over a specified time period, you can include the duration and resolution of this temporal coverage.
    """
    # Supplying duration is optional but both fields within it required
    duration: Optional[TemporalDurationInfo] = None

    # Temporal resolution - ISO8601 Duration format e.g. "P1Y2M10DT2H30M"
    resolution: Optional[str] = Field(
        None,
        title="Temporal Resolution",
        description="The temporal resolution (i.e. time step) of the data. Please use the [ISO8601 duration format](https://en.wikipedia.org/wiki/ISO_8601#Durations) e.g. \"P1Y2M10DT2H30M\".",
    )

    # validate the temporal resolution conforms to expected format: "P1Y2M10DT2H30M\".
    # TODO[pydantic]: We couldn't refactor the `validator`, please replace it by `field_validator` manually.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-validators for more information.
    @validator('resolution', always=True)  # run even for default value.
    def validate_temporal_resolution(cls: Any, v: str) -> str:
        try:
            if v is not None:
                parse_duration(v)
        except Exception as e:
            raise ValueError(
                f"Invalid temporal resolution. The value must conform to the ISO8601 Time Duration format (e.g.'P1Y2M10DT2H30M'). Details: {e}")
        return v
    model_config = ConfigDict(
        str_min_length=1, extra="forbid", title="Dataset Temporal Information")


class OptionallyRequiredDate(BaseModel):
    relevant: bool = False
    value: Optional[date] = None

    @model_validator(mode='after')
    def check_value_provided(self) -> Self:
        relevant: Optional[bool] = self.relevant
        value: Optional[date] = self.value

        assert relevant is not None

        if relevant == True:
            if value is None:
                raise ValueError(
                    "Must provide a date if the relevant field is set to True.")
        else:
            if value is not None:
                raise ValueError(
                    "Cannot provide a date if the relevant field is set to False.")
        return self


class PublishedDate(OptionallyRequiredDate):
    """
    Has the dataset been published? If so, please provide the date on which this version of the dataset was first published.
    """
    model_config = ConfigDict(extra="forbid", title="Dataset Publication Date")


class CreatedDate(OptionallyRequiredDate):
    """
    Has the dataset been created? If so, please provide the date on which this version of the dataset was produced or generated.
    """
    model_config = ConfigDict(extra="forbid", title="Dataset Creation Date")


class CollectionFormatDatasetInfo(BaseModel):
    """
    Please provide information about this dataset, including the name, description, creation date and publisher information.
    """
    name: str = Field(
        ...,
        title="Dataset name",
        description="Please provide a readable name for the dataset.",
    )
    description: str = Field(
        ...,
        title="Description",
        description="Please provide a more detailed description of the dataset. This should include the nature of the data, the intended usage, and any other relevant information.",
    )

    # information about whether the dataset is reposited or not
    access_info: AccessInfo

    # linked publisher in registry
    publisher_id: str = Field(
        ...,
        title="Publisher",
        description="Please provide information about the organisation which is publishing or producing this dataset. If this is your organisation, please select it again using the tool below.",
    )

    # created/published dates
    created_date: CreatedDate

    published_date: PublishedDate

    # License field - selected by drop down
    license: AnyHttpUrl = Field(
        title="Usage license",
        description="Please select a standard license for usage, by default it will be 'Copyright'.",
        json_schema_extra={"format": "uri"}
    )

    # What is the purpose of this dataset?
    purpose: Optional[str] = Field(
        None,
        title="Dataset Purpose",
        description="A brief description of the reason a data asset was created. Should be a good guide to the potential usefulness of a data asset to other users.",
    )

    # Who holds the rights to this dataset? Free text input
    rights_holder: Optional[str] = Field(
        None,
        title="Dataset Rights Holder",
        description="Specify the party owning or managing rights over the resource. Please ensure you have sought consent to include these details in the record.",
    )

    # Usage limitations? Free text input
    usage_limitations: Optional[str] = Field(
        None,
        title="Usage Limitations",
        description="A statement that provides information on any caveats or restrictions on access or on the use of the data asset, including legal, security, privacy, commercial or other limitations.",
    )

    preferred_citation: Optional[str] = Field(
        None,
        title="Preferred dataset citation",
        description="If you would like to specify how this dataset should be cited, please use the checkbox below to indicate this preference, and enter your preferred citation in the text box.",
    )

    # Spatial information?
    spatial_info: Optional[CollectionFormatSpatialInfo] = None

    # Temporal information?
    temporal_info: Optional[CollectionFormatTemporalInfo] = None

    # What formats are included in the dataset files e.g. txt, pdf, html, css, json etc
    formats: Optional[List[str]] = Field(
        None,
        title="Dataset File Formats",
        description="What file formats are present in this dataset? E.g. \"pdf\", \"csv\" etc.",
    )

    # Keywords associated with this dataset

    # TODO validate this JSON schema is suitable
    keywords: Optional[List[str]] = Field(
        None,
        title="Keywords",
        description="Provide a list of keywords which describe your dataset [Optional].",
    )

    # Custom user metadata if desired - this field is kept synchronised with the
    # base domain info user metadata TODO consider how we can more naturally
    # include move the dataset collection format into the domain info while
    # allowing for immutable fields/properties such as s3 location etc
    user_metadata: Optional[Dict[str, str]] = Field(
        None, title="Custom User Metadata", description="Optionally provide a collection of key value annotations for this dataset.")

    # this is not going to be shown in schema anymore, but will remain in the
    # actual model in case existing versions are relevant
    version: Optional[str] = None
    model_config = ConfigDict(
        str_min_length=1, extra="forbid", title="Dataset Information")


class CollectionFormatApprovals(BaseModel):
    """
    Please ensure that the following approvals are considered. Your dataset may not be fit for inclusion in the Data Store if it does not meet the following requirements.
    """
    ethics_registration: DatasetEthicsRegistrationCheck
    ethics_access: DatasetEthicsAccessCheck
    indigenous_knowledge: IndigenousKnowledgeCheck
    export_controls: ExportControls
    model_config = ConfigDict(
        str_min_length=1, extra="forbid", title="Dataset Approvals")


class CollectionFormat(BaseModel):
    """
    This form collects metadata about the dataset you are registering as well as the registration itself. Please ensure the fields below are completed accurately.
    """
    # Required
    associations: CollectionFormatAssociations
    approvals: CollectionFormatApprovals
    dataset_info: CollectionFormatDatasetInfo
    model_config = ConfigDict(
        str_min_length=1, extra="forbid", title="Dataset Metadata")


class S3Location(BaseModel):
    bucket_name: str
    path: str
    s3_uri: str


class ReleasedStatus(str, Enum):
    # If changing this, also update QueryDatasetReleaseStatusType in RegistryAPI
    # which should mirror this but python doesnt allow for inheriting enum classes.

    # Aprroval status for release of datasets
    NOT_RELEASED = "NOT_RELEASED"
    PENDING = "PENDING"
    RELEASED = "RELEASED"


class ReleaseAction(str, Enum):
    # Action taken on a dataset regarding releasing
    REQUEST = "REQUEST"
    APPROVE = "APPROVE"
    REJECT = "REJECT"


class ReleaseHistoryEntry(BaseModel):
    # What action was taken
    action: ReleaseAction

    # What time did this occur?
    timestamp: int

    # All actions relate to an approver
    approver: IdentifiedResource

    # Some actions relate to a requester - i.e. Requesting
    requester: Optional[IdentifiedResource] = None

    # All actions are associated with some notes/details
    notes: str


class DatasetDomainInfo(DomainInfoBase):
    # main dataset metadata that data store front end collects
    collection_format: CollectionFormat

    # the s3 location information NOTE This is now optional as not all
    # registered datasets allow data repositing
    s3: S3Location

    release_history: List[ReleaseHistoryEntry] = []

    # need here for Global Secondary Index to search by approver.
    release_status: ReleasedStatus
    release_approver: Optional[IdentifiedResource] = None
    # timestamp when release status was last updated
    release_timestamp: Optional[int] = None
    # collection_format.access_info.uri for non-reposited datasets to enable fetch/list by uri.
    access_info_uri: Optional[AnyUrl] = Field(
        default=None,
        title="Access Info URI",
        description="Provide the best URI of the externally hosted data.",
    )

    # validate access_info_uri and collection_format.dataset_info.access_info.uri are always the same
    @model_validator(mode='after')
    def check_access_info_uri(self) -> Self:
        assert self.collection_format, "collection_format must be present"
        access_info = self.collection_format.dataset_info.access_info
        assert (access_info.uri and self.access_info_uri) or (
            not access_info.uri and not self.access_info_uri), f"access_info_uri must always match collection_format.access_info.uri. Got access_info_uri={self.access_info_uri} and collection_format.access_info.uri={access_info.uri}"
        if str(self.collection_format.dataset_info.access_info.uri) != str(self.access_info_uri):
            raise ValueError(
                f"access_info_uri must always match collection_format.access_info.uri. Got access_info_uri={self.access_info_uri} and collection_format.access_info.uri={access_info.uri}")
        return self

    # validate release_approver and release_timestamp are both present or both absent
    @model_validator(mode='after')
    def check_release_status(self) -> Self:
        # make sure both fields are present or both are absent
        if (self.release_approver is None) != (self.release_timestamp is None):
            raise ValueError(
                f"Cannot provide a release activity timestamp without release approver and vice versa. Got release_approver={self.release_approver} and release_timestamp={self.release_timestamp}")
        return self


class ItemDataset(EntityBase, DatasetDomainInfo):
    # override sub type
    item_subtype: ItemSubType = ItemSubType.DATASET

    # override history
    history: List[HistoryEntry[DatasetDomainInfo]]  # type: ignore

    @staticmethod
    def get_searchable_fields() -> List[str]:
        # this is a special elastic search syntax
        return ItemBase.get_searchable_fields() + [
            'description',
            'name',
            'publisher',
            'organisation',
            'keywords^2',
            'formats',
            'preferred_citation',
            'usage_limitations',
            'rights_holder',
            'purpose',
            'data_custodian'
        ]

    def get_search_ready_object(self) -> Dict[str, str]:
        # get base info
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()

        # pull out relevant sub fields
        cf = self.collection_format
        di: CollectionFormatDatasetInfo = cf.dataset_info
        assoc: CollectionFormatAssociations = cf.associations
        org = assoc.organisation_id

        extended: Dict[str, str] = {
            'description': di.description,
            'name': di.name,
            'publisher': di.publisher_id,
            'organisation': org,
            'keywords': " ".join(di.keywords or []),
            'formats': " ".join(di.formats or []),
            'preferred_citation': di.preferred_citation or "",
            'usage_limitations': di.usage_limitations or "",
            'rights_holder': di.rights_holder or "",
            'purpose': di.purpose or "",
            'data_custodian': assoc.data_custodian_id or ""
        }
        base.update(extended)

        return base

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v

# ======================================================================================
# Model Workflow Template (Entity - Workflow Template <- Model Workflow Template)
# ======================================================================================


class ModelRunWorkflowTemplateDomainInfo(WorkflowTemplateDomainInfo):
    # expects that the software references a Model rather than just generic software
    # this can be enforced at ingestion time
    None


class ItemModelRunWorkflowTemplate(EntityBase, ModelRunWorkflowTemplateDomainInfo):
    # override sub type
    item_subtype: ItemSubType = ItemSubType.MODEL_RUN_WORKFLOW_TEMPLATE

    # override history
    history: List[  # type: ignore
        HistoryEntry[ModelRunWorkflowTemplateDomainInfo]]

    def get_search_ready_object(self) -> Dict[str, str]:
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
            "software_id": self.software_id,
            "input_templates": " ".join([t.template_id for t in self.input_templates]),
            "output_templates": " ".join([t.template_id for t in self.output_templates]),
            "annotations": self.annotations.make_searchable() if self.annotations else ""
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        return ItemBase.get_searchable_fields() + [
            "software_id",
            "input_templates",
            "output_templates",
            "annotations"
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v


"""
==========
ACTIVITIES
==========
"""

# model run

# avoid circular imports by importing late
try:
    from ProvenaInterfaces.ProvenanceModels import ModelRunRecord
except:
    from .ProvenanceModels import ModelRunRecord


class WorkflowRunCompletionStatus(str, Enum):
    INCOMPLETE = "INCOMPLETE"
    COMPLETE = "COMPLETE"
    LODGED = "LODGED"

# =========================
# Study (Activity <- Study)
# =========================


class StudyDomainInfo(DomainInfoBase):
    title: str
    description: str
    study_alternative_id: Optional[str] = None


class ItemStudy(ActivityBase, StudyDomainInfo):
    item_subtype: ItemSubType = ItemSubType.STUDY

    # override history
    history: List[HistoryEntry[StudyDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        # no extra searchable info - just return parent
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
            'title': self.title,
            'description': self.title,
            'study_alternative_id': self.study_alternative_id or "",
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        # no extra searchable info - just return parent
        return ItemBase.get_searchable_fields() + [
            'title',
            'description',
            'study_alternative_id'
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v


# ==========================================
# Registry Version (Activity <- Registry Version)
# ==========================================


class VersionDomainInfo(DomainInfoBase):
    # Should include link to the created item
    reason: str
    from_item_id: str
    to_item_id: str
    new_version_number: int


class ItemVersion(ActivityBase, VersionDomainInfo):
    item_subtype: ItemSubType = ItemSubType.VERSION

    # override history
    history: List[HistoryEntry[VersionDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        # no extra searchable info - just return parent
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
            'from_item_id': self.from_item_id,
            'to_item_id': self.to_item_id,
            'new_version_number': str(self.new_version_number)
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        # no extra searchable info - just return parent
        return ItemBase.get_searchable_fields() + [
            'from_item_id',
            'to_item_id',
            'new_version_number'
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v


# ==========================================
# Registry Create (Activity <- Workflow Run)
# ==========================================


class CreateDomainInfo(DomainInfoBase):
    # The base info includes other important info e.g. person, time etc

    # Should include link to the created item
    created_item_id: str


class ItemCreate(ActivityBase, CreateDomainInfo):
    item_subtype: ItemSubType = ItemSubType.CREATE

    # override history
    history: List[HistoryEntry[CreateDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        # no extra searchable info - just return parent
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
            "created_item_id": self.created_item_id
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        # no extra searchable info - just return parent
        return ItemBase.get_searchable_fields() + [
            "created_item_id"
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v

# ========================================
# Workflow Run (Activity - Workflow Run)
# ========================================


class WorkflowRunDomainInfo(DomainInfoBase):
    # status of the record
    record_status: WorkflowRunCompletionStatus
    None


class ItemWorkflowRun(ActivityBase, WorkflowRunDomainInfo):
    item_subtype: ItemSubType = ItemSubType.WORKFLOW_RUN

    # override history
    history: List[HistoryEntry[WorkflowRunDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        # no extra searchable info - just return parent
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        # no extra searchable info - just return parent
        return ItemBase.get_searchable_fields() + [
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v


# =================================================
# Model Run (Activity - Workflow Run <- Model Run)
# =================================================


class ModelRunDomainInfo(WorkflowRunDomainInfo):
    # original input format
    record: ModelRunRecord
    # serialised prov document
    prov_serialisation: str


class ItemModelRun(ActivityBase, ModelRunDomainInfo):
    # override sub type
    item_subtype: ItemSubType = ItemSubType.MODEL_RUN

    # override history
    history: List[HistoryEntry[ModelRunDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()

        # base info
        extended: Dict[str, str] = {
            "prov_serialisation": self.prov_serialisation,
        }

        # and extend based on model run records make searchable function
        model_run_record_search_object = self.record.get_search_ready_object()
        extended.update(model_run_record_search_object)
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        item_base = ItemBase.get_searchable_fields()
        model_run_record_fields = ModelRunRecord.get_searchable_fields()
        above_fields = [
            "prov_serialisation"
        ]
        return item_base + model_run_record_fields + above_fields

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v


"""
========
AGENTS
========
"""

# ========================
# Person (Agent - Person)
# ========================


class PersonDomainInfo(DomainInfoBase):
    email: EmailStr

    # person info
    first_name: str
    last_name: str

    # optional info
    orcid: Optional[AnyHttpUrl] = None

    # Consent to make this person
    ethics_approved: bool = False


class ItemPerson(AgentBase, PersonDomainInfo):
    # override sub type
    item_subtype: ItemSubType = ItemSubType.PERSON

    # override history
    history: List[HistoryEntry[PersonDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
            "first_name": self.first_name,
            "last_name": self.first_name,
            "email": self.email,
            "orcid": str(self.orcid) if self.orcid else ""
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        return ItemBase.get_searchable_fields() + [
            "first_name",
            "last_name",
            "email",
            "orcid"
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v

# ====================================
# Organisation (Agent - Organisation)
# ====================================


class OrganisationDomainInfo(DomainInfoBase):
    # organisation info
    name: str

    # optional info
    ror: Optional[AnyHttpUrl] = None


class ItemOrganisation(AgentBase, OrganisationDomainInfo):
    # override sub type
    item_subtype: ItemSubType = ItemSubType.ORGANISATION

    # override history
    history: List[HistoryEntry[OrganisationDomainInfo]]  # type: ignore

    def get_search_ready_object(self) -> Dict[str, str]:
        item_base = ItemBase.parse_obj(self.dict())
        base = item_base.get_search_ready_object()
        extended: Dict[str, str] = {
            "name": self.name,
            "ror": str(self.ror) if self.ror else ""
        }
        base.update(extended)
        return base

    @staticmethod
    def get_searchable_fields() -> List[str]:
        return ItemBase.get_searchable_fields() + [
            "name",
            "ror"
        ]

    @field_validator('history')
    @classmethod
    def unique_history(cls: Any, v: List[HistoryEntry[Any]]) -> Any:
        # validate that the ids are unique in the history
        unique_history_ids(v)
        return v


MODEL_TYPE_MAP: Dict[Tuple[ItemCategory, ItemSubType], Type[ItemBase]] = {
    (ItemCategory.ACTIVITY, ItemSubType.MODEL_RUN): ItemModelRun,
    (ItemCategory.ACTIVITY, ItemSubType.STUDY): ItemStudy,
    (ItemCategory.ACTIVITY, ItemSubType.CREATE): ItemCreate,
    (ItemCategory.ACTIVITY, ItemSubType.VERSION): ItemVersion,
    (ItemCategory.AGENT, ItemSubType.PERSON): ItemPerson,
    (ItemCategory.AGENT, ItemSubType.ORGANISATION): ItemOrganisation,
    (ItemCategory.ENTITY, ItemSubType.MODEL): ItemModel,
    (ItemCategory.ENTITY, ItemSubType.MODEL_RUN_WORKFLOW_TEMPLATE): ItemModelRunWorkflowTemplate,
    (ItemCategory.ENTITY, ItemSubType.DATASET_TEMPLATE): ItemDatasetTemplate,
    (ItemCategory.ENTITY, ItemSubType.DATASET): ItemDataset,
}
MODEL_DOMAIN_INFO_TYPE_MAP: Dict[Tuple[ItemCategory, ItemSubType], Type[DomainInfoBase]] = {
    (ItemCategory.ACTIVITY, ItemSubType.STUDY): StudyDomainInfo,
    (ItemCategory.ACTIVITY, ItemSubType.MODEL_RUN): ModelRunDomainInfo,
    (ItemCategory.ACTIVITY, ItemSubType.CREATE): CreateDomainInfo,
    (ItemCategory.ACTIVITY, ItemSubType.VERSION): VersionDomainInfo,
    (ItemCategory.AGENT, ItemSubType.PERSON): PersonDomainInfo,
    (ItemCategory.AGENT, ItemSubType.ORGANISATION): OrganisationDomainInfo,
    (ItemCategory.ENTITY, ItemSubType.MODEL): ModelDomainInfo,
    (ItemCategory.ENTITY, ItemSubType.MODEL_RUN_WORKFLOW_TEMPLATE): ModelRunWorkflowTemplateDomainInfo,
    (ItemCategory.ENTITY, ItemSubType.DATASET_TEMPLATE): DatasetTemplateDomainInfo,
    (ItemCategory.ENTITY, ItemSubType.DATASET): DatasetDomainInfo,
}


# create a merged collection of all unique searchable fields of registry models
# as per above map
ALL_SEARCHABLE_FIELDS = list(set(
    itertools.chain(*[model_type.get_searchable_fields()
                    for model_type in MODEL_TYPE_MAP.values()])
))



from ProvenaInterfaces.RegistryModels import *
from tests.helpers.helpers import *
from ProvenaInterfaces.TestConfig import route_params, RouteParameters
import pytest
# Ignore mypy errors as we are testing invalid models here and mypy
# doesn't know to look up and see "with self.assertRaises(ValueError):"
# The correct models will still fail if they aren't actually correct in the success case test
# mypy: ignore-errors


def test_spatial_info() -> None:
    spatial_info = CollectionFormatSpatialInfo(resolution="1.23")
    # With validation error:
    with pytest.raises(ValueError):
        spatial_info = CollectionFormatSpatialInfo(resolution="1.23abcd")


def test_temporal_info() -> None:
    # Test success cases
    temporal_info = CollectionFormatTemporalInfo(
        duration=TemporalDurationInfo(
            begin_date="2019-01-01",
            end_date="2019-01-02"
        ),
        resolution="P1Y2M10DT2H30M"
    )

    # ensure failure with missing begin or end date
    with pytest.raises(ValueError):
        temporal_info = CollectionFormatTemporalInfo(
            duration=TemporalDurationInfo(
                begin_date="2019-01-01"
            ),
            resolution="P1Y2M10DT2H30M"
        )

    with pytest.raises(ValueError):
        temporal_info = CollectionFormatTemporalInfo(
            duration=TemporalDurationInfo(
                end_date="2019-01-01"
            ),
            resolution="P1Y2M10DT2H30M"
        )

    # ensure failure with invalid resolution
    with pytest.raises(ValueError):
        temporal_info = CollectionFormatTemporalInfo(
            duration=TemporalDurationInfo(
                begin_date="2019-01-01",
                end_date="2019-01-02"
            ),
            resolution="P1Y2M10DT2H30Mabcd"
        )

    # ensure failure with invalid order of start and end date
    with pytest.raises(ValueError):
        temporal_info = CollectionFormatTemporalInfo(
            duration=TemporalDurationInfo(
                begin_date="2019-01-02",
                end_date="2019-01-01"
            ),
            resolution="P1Y2M10DT2H30M"
        )

    # ensure success with None for temporal resolution
    temporal_info = CollectionFormatTemporalInfo(
        duration=TemporalDurationInfo(
            begin_date="2019-01-01",
            end_date="2019-01-02"
        ),
        resolution=None
    )


def test_dataset_access_info_uri_set() -> None:
    # Test that access_info_uri is set correctly
    dataset: DatasetDomainInfo = get_model_example(
        item_subtype=ItemSubType.DATASET)
    assert dataset.access_info_uri == None  # check the same for the None case
    assert dataset.access_info_uri == dataset.collection_format.dataset_info.access_info.uri

    dataset.access_info_uri = "https://example.com"
    dataset.collection_format.dataset_info.access_info.uri = "https://example.com"
    dataset.collection_format.dataset_info.access_info.reposited = False
    dataset.collection_format.dataset_info.access_info.description = 'Test'
    # re parse to run validators
    dataset = DatasetDomainInfo.parse_obj(dataset.dict())

    # Test that ValueError is raised when access_info_uri is not set
    dataset.access_info_uri = None
    assert dataset.collection_format.dataset_info.access_info.uri
    with pytest.raises(ValueError):
        dataset = DatasetDomainInfo.parse_obj(dataset.dict())

    # Test if dataset.access_info_uri is set but collection_format.dataset_info.access_info.uri is not
    with pytest.raises(ValueError):
        dataset.access_info_uri = "https://example.com"
        dataset.collection_format.dataset_info.access_info.uri = None
        dataset = DatasetDomainInfo.parse_obj(dataset.dict())

    # Test that ValueError is raised when access_info_uri does not match collection_format.access_info.uri
    with pytest.raises(ValueError):
        dataset.collection_format.dataset_info.access_info.uri = dataset.access_info_uri + "/abcd"
        dataset = DatasetDomainInfo.parse_obj(dataset.dict())


# Test for PublicationDate
def test_publication_date_validation():
    # Test case where relevant is True and a date is provided
    pub_date = PublishedDate(relevant=True, value=date(2020, 1, 1))
    assert pub_date.value == date(2020, 1, 1)

    # Test case where relevant is True but no date is provided
    with pytest.raises(ValueError) as excinfo:
        PublishedDate(relevant=True)
    assert "Must provide a date if the relevant field is set to True" in str(
        excinfo.value)

    # Test case where relevant is False and no date is provided
    pub_date = PublishedDate(relevant=False)
    assert pub_date.value is None

    # Test case where relevant is False but a date is provided
    with pytest.raises(ValueError) as excinfo:
        PublishedDate(relevant=False, value=date(2020, 1, 1))
    assert "Cannot provide a date if the relevant field is set to False" in str(
        excinfo.value)

# Test for CreatedDate


def test_creation_date_validation():
    # Test case where relevant is True and a date is provided
    creation_date = CreatedDate(relevant=True, value=date(2020, 1, 1))
    assert creation_date.value == date(2020, 1, 1)

    # Test case where relevant is True but no date is provided
    with pytest.raises(ValueError) as excinfo:
        CreatedDate(relevant=True)
    assert "Must provide a date if the relevant field is set to True" in str(
        excinfo.value)

    # Test case where relevant is False and no date is provided
    creation_date = CreatedDate(relevant=False)
    assert creation_date.value is None

    # Test case where relevant is False but a date is provided
    with pytest.raises(ValueError) as excinfo:
        CreatedDate(relevant=False, value=date(2020, 1, 1))
    assert "Cannot provide a date if the relevant field is set to False" in str(
        excinfo.value)



from typing import TypeVar, Generic, Type
from ProvenaInterfaces.RegistryAPI import *
from ProvenaInterfaces.ProvenanceModels import *
from dataclasses import dataclass
from datetime import datetime

# Type variables to describe generic parent classes
SeedResponseTypeVar = TypeVar(
    'SeedResponseTypeVar', bound=Union[None, GenericSeedResponse])
FetchResponseTypeVar = TypeVar(
    'FetchResponseTypeVar', bound=GenericFetchResponse)
ListResponseTypeVar = TypeVar(
    'ListResponseTypeVar', bound=GenericListResponse)
CreateResponseTypeVar = TypeVar(
    'CreateResponseTypeVar', bound=Union[None, GenericCreateResponse])
ItemModelTypeVar = TypeVar('ItemModelTypeVar', bound=ItemBase)
ItemDomainInfoTypeVar = TypeVar('ItemDomainInfoTypeVar', bound=DomainInfoBase)

# Generic based data class for containing
# typing classes


@dataclass
class TypeParameters(Generic[
    ItemDomainInfoTypeVar,
    ItemModelTypeVar,
    SeedResponseTypeVar,
    FetchResponseTypeVar,
    ListResponseTypeVar,
    CreateResponseTypeVar
]):
    domain_info: Type[ItemDomainInfoTypeVar]
    item_model: Type[ItemModelTypeVar]
    seed_response: Optional[Type[SeedResponseTypeVar]]
    fetch_response: Type[FetchResponseTypeVar]
    list_response: Type[ListResponseTypeVar]
    create_response: Optional[Type[CreateResponseTypeVar]]

# Class containing example models for each required type
# e.g. create requires example domain info types


@dataclass
class ModelExamples(Generic[ItemDomainInfoTypeVar]):
    domain_info: List[ItemDomainInfoTypeVar]

# Collection of parameters for each test


@dataclass
class RouteParameters(Generic[
    ItemDomainInfoTypeVar,
    ItemModelTypeVar,
    SeedResponseTypeVar,
    FetchResponseTypeVar,
    ListResponseTypeVar,
    CreateResponseTypeVar
]):
    route: str
    name: str

    # this is required for dataset/model run which only have proxy variant of
    # seed/create/update routes and require the special roles
    category: ItemCategory
    subtype: ItemSubType

    typing_information: TypeParameters[
        ItemDomainInfoTypeVar,
        ItemModelTypeVar,
        SeedResponseTypeVar,
        FetchResponseTypeVar,
        ListResponseTypeVar,
        CreateResponseTypeVar
    ]
    model_examples: ModelExamples[
        ItemDomainInfoTypeVar
    ]

    use_special_proxy_modify_routes: bool = False


route_params: List[RouteParameters] = [
    # ACTIVITY - MODEL RUN
    RouteParameters(
        use_special_proxy_modify_routes=True,
        route="/activity/model_run",
        name="Activity - Model Run",
        category=ItemCategory.ACTIVITY,
        subtype=ItemSubType.MODEL_RUN,
        typing_information=TypeParameters[
            ModelRunDomainInfo,
            ItemModelRun,
            ModelRunSeedResponse,
            ModelRunFetchResponse,
            ModelRunListResponse,
            ModelRunCreateResponse
        ](
            domain_info=ModelRunDomainInfo,
            item_model=ItemModelRun,
            seed_response=ModelRunSeedResponse,
            fetch_response=ModelRunFetchResponse,
            list_response=ModelRunListResponse,
            create_response=ModelRunCreateResponse
        ),
        model_examples=ModelExamples[ModelRunDomainInfo](
            domain_info=[
                ModelRunDomainInfo(
                    display_name="Example model run",
                    prov_serialisation="Blank prov document",
                    record_status=WorkflowRunCompletionStatus.COMPLETE,
                    user_metadata=None,
                    record=ModelRunRecord(
                        model_version="1.2",
                        workflow_template_id="1234",
                        inputs=[TemplatedDataset(
                            dataset_template_id="1234",
                            dataset_id="1234",
                            dataset_type=DatasetType.DATA_STORE,
                            resources={
                                '1234': '/path/to/resource.csv'
                            }
                        )],
                        outputs=[TemplatedDataset(
                            dataset_template_id="1234",
                            dataset_id="1234",
                            dataset_type=DatasetType.DATA_STORE,
                            resources={
                                '1234': '/path/to/resource.csv'
                            }
                        )],
                        associations=AssociationInfo(
                            modeller_id="3247uy82",
                            requesting_organisation_id="58493058"
                        ),
                        start_time=int((datetime.now().timestamp())),
                        end_time=int((datetime.now().timestamp())),
                        display_name="Example model run",
                        description="This is a fake model run",
                        annotations={
                            'somekey': 'somevalue'
                        }
                    )
                ),
                ModelRunDomainInfo(
                    display_name="Example model run 2",
                    prov_serialisation="Blank prov document 2",
                    record_status=WorkflowRunCompletionStatus.COMPLETE,
                    user_metadata=None,
                    record=ModelRunRecord(
                        workflow_template_id="12345",
                        inputs=[TemplatedDataset(
                            dataset_template_id="12354",
                            dataset_id="15234",
                            dataset_type=DatasetType.DATA_STORE,
                            resources={
                                '12534': '/path/to/resource.csv'
                            }
                        )],
                        outputs=[],
                        associations=AssociationInfo(
                            modeller_id="324dsjfilks7uy82",
                            requesting_organisation_id="584jdlsf93058"
                        ),
                        display_name="Example model run 2",
                        description="This is a fake model run 2",
                        start_time=int((datetime.now().timestamp())),
                        end_time=int((datetime.now().timestamp())),
                    )
                )
            ]

        )
    ),
    # ENTITY - DATASET
    RouteParameters(
        use_special_proxy_modify_routes=True,
        route="/entity/dataset",
        name="Entity - Dataset",
        category=ItemCategory.ENTITY,
        subtype=ItemSubType.DATASET,
        typing_information=TypeParameters[
            DatasetDomainInfo,
            ItemDataset,
            DatasetSeedResponse,
            DatasetFetchResponse,
            DatasetListResponse,
            DatasetCreateResponse
        ](
            domain_info=DatasetDomainInfo,
            item_model=ItemDataset,
            seed_response=DatasetSeedResponse,
            fetch_response=DatasetFetchResponse,
            list_response=DatasetListResponse,
            create_response=DatasetCreateResponse
        ),
        model_examples=ModelExamples[DatasetDomainInfo](
            domain_info=[
                DatasetDomainInfo(
                    display_name="example dataset 1",
                    release_approver=None,
                    release_timestamp=None,
                    access_info_uri=None,
                    collection_format=CollectionFormat(
                        associations=CollectionFormatAssociations(
                            organisation_id="1234",
                            data_custodian_id=None,
                            point_of_contact="Not Peter Baker.",
                        ),
                        approvals=CollectionFormatApprovals(
                            ethics_registration=DatasetEthicsRegistrationCheck(),
                            ethics_access=DatasetEthicsAccessCheck(),
                            indigenous_knowledge=IndigenousKnowledgeCheck(
                            ),
                            export_controls=ExportControls(
                            )
                        ),
                        dataset_info=CollectionFormatDatasetInfo(
                            name='test dataset 1',
                            description='test dataset 1',
                            access_info=AccessInfo(
                                reposited=True, uri=None, description=None),
                            version=None,
                            publisher_id="1234",
                            created_date=CreatedDate(
                                relevant=True,
                                value=date.today()
                            ),
                            published_date=PublishedDate(
                                relevant=True,
                                value=date.today()
                            ),
                            license="https://example.com",
                            keywords=["a", "b"],
                            spatial_info=CollectionFormatSpatialInfo(
                                coverage="SRID=4326;POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))",
                                resolution="0.1",
                                extent="SRID=4326;POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"
                            ),
                            temporal_info=CollectionFormatTemporalInfo(
                                duration=TemporalDurationInfo(
                                    begin_date=date.today(),
                                    end_date=date.today(),
                                ),
                                resolution="P1Y2M10DT2H30M"
                            ),
                            purpose="But why, you might ask, was the Test Dataset so important? Well, dear reader, it served as a mirror reflecting the very essence of the software, exposing its vulnerabilities and frailties. It was Testy's trustworthy sidekick, always ready to point out flaws, inconsistencies, and missteps in the code. It was the guardian of reliability, ensuring that the software would perform its duties diligently without crashing or misbehaving. And it was the staunch defender of predictability, making sure that the software would respond consistently to various inputs and scenarios.",
                            usage_limitations="You may not use this dataset while wearing an eyepatch, unless you are an actual pirate sailing the high seas. In that case, a parrot on your shoulder is also required for compliance.",
                            preferred_citation="Testy McTestface, \"The Test Dataset: A Mirror to the Soul of the Software\", Journal of Testy Software, 2021",
                            formats=["pdf"],
                            rights_holder="ME",
                            user_metadata={
                                "my custom": "annotation",
                                "another custom": "annotation"
                            }
                        )
                    ),
                    s3=S3Location(
                        bucket_name="example",
                        path="example",
                        s3_uri="example"
                    ),
                    release_status=ReleasedStatus.NOT_RELEASED,
                    user_metadata={
                        "my custom": "annotation",
                        "another custom": "annotation"
                    }
                ),
                DatasetDomainInfo(
                    display_name="example dataset 2",
                    release_approver=None,
                    release_timestamp=None,
                    collection_format=CollectionFormat(
                        associations=CollectionFormatAssociations(
                            organisation_id="12345",
                            data_custodian_id=None,
                            point_of_contact=None
                        ),
                        approvals=CollectionFormatApprovals(
                            ethics_registration=DatasetEthicsRegistrationCheck(),
                            ethics_access=DatasetEthicsAccessCheck(),
                            indigenous_knowledge=IndigenousKnowledgeCheck(
                            ),
                            export_controls=ExportControls(
                            )
                        ),
                        dataset_info=CollectionFormatDatasetInfo(
                            name='test dataset 2',
                            rights_holder=None,
                            preferred_citation=None,
                            keywords=None,
                            version=None,
                            description='test dataset 2',
                            access_info=AccessInfo(
                                reposited=False,
                                uri="https://fake.url.com/resource",
                                description="You can access this fake dataset by doing nothing at all!"
                            ),
                            publisher_id="1234",
                            created_date=CreatedDate(
                                relevant=True,
                                value=date.today()
                            ),
                            published_date=PublishedDate(
                                relevant=True,
                                value=date.today()
                            ),
                            license="https://example.com",
                            spatial_info=CollectionFormatSpatialInfo(
                                coverage="SRID=4326;POLYGON ((20 10, 40 40, 20 40, 10 20, 30 10))",
                                resolution="1.23456",
                                extent=None
                            ),
                            temporal_info=CollectionFormatTemporalInfo(
                                duration=TemporalDurationInfo(
                                    begin_date=date.today(),
                                    end_date=date.today(),
                                ),
                                resolution="P1Y2M10DT2H24M"
                            ),
                            purpose="Explore the intricacies of intergalactic communication with this dataset, which deciphers the meaning behind alien emojis. Learn when it's appropriate to send a cosmic winky face or a UFO thumbs-up in your next close encounter.",
                            usage_limitations=" If you accidentally drop your device while using the dataset, you must perform a ceremonial dance to appease the digital spirits and ensure your files remain unharmed.",
                            formats=["png"],
                            user_metadata={
                                "my updated custom": "annotation",
                            }
                        ),
                    ),
                    s3=S3Location(
                        bucket_name="example",
                        path="example",
                        s3_uri="example"
                    ),
                    release_status=ReleasedStatus.NOT_RELEASED,
                    access_info_uri="https://fake.url.com/resource",
                    user_metadata={
                        "my updated custom": "annotation",
                    }
                ),
            ]

        )
    ),
    # AGENT - ORGANISATION
    RouteParameters(
        route="/agent/organisation",
        name="Agent - Organisation",
        category=ItemCategory.AGENT,
        subtype=ItemSubType.ORGANISATION,
        typing_information=TypeParameters[
            OrganisationDomainInfo,
            ItemOrganisation,
            OrganisationSeedResponse,
            OrganisationFetchResponse,
            OrganisationListResponse,
            OrganisationCreateResponse
        ](
            domain_info=OrganisationDomainInfo,
            item_model=ItemOrganisation,
            seed_response=OrganisationSeedResponse,
            fetch_response=OrganisationFetchResponse,
            list_response=OrganisationListResponse,
            create_response=OrganisationCreateResponse
        ),
        model_examples=ModelExamples[OrganisationDomainInfo](
            domain_info=[
                OrganisationDomainInfo(
                    display_name="Test org",
                    name="Test org",
                    ror="http://example.org/test-org",
                    user_metadata={
                        "my custom": "annotation",
                        "another custom": "annotation"
                    }
                ),
                OrganisationDomainInfo(
                    display_name="Test org updated",
                    name="Test org updated",
                    ror="http://example.org/test-org",
                    user_metadata={
                        "my updated custom": "annotation",
                    }
                )
            ]
        )
    ),

    # AGENT - PERSON
    RouteParameters(
        route="/agent/person",
        name="Agent - Person",
        category=ItemCategory.AGENT,
        subtype=ItemSubType.PERSON,
        typing_information=TypeParameters[
            PersonDomainInfo,
            ItemPerson,
            PersonSeedResponse,
            PersonFetchResponse,
            PersonListResponse,
            PersonCreateResponse
        ](
            domain_info=PersonDomainInfo,
            item_model=ItemPerson,
            seed_response=PersonSeedResponse,
            fetch_response=PersonFetchResponse,
            list_response=PersonListResponse,
            create_response=PersonCreateResponse
        ),
        model_examples=ModelExamples[PersonDomainInfo](
            domain_info=[
                PersonDomainInfo(
                    display_name="Le Test bot",
                    first_name="Test",
                    last_name="Boy",
                    email="test@example.com",
                    orcid="https://orcid.org/0000-1234-1234-1234",
                    ethics_approved=True,
                    user_metadata={
                        "my custom": "annotation",
                        "another custom": "annotation"
                    }
                ),
                PersonDomainInfo(
                    display_name="Le Test bot updated",
                    first_name="Test",
                    last_name="Man",
                    email="test@example.com",
                    orcid="https://orcid.org/0000-1234-1234-1234",
                    ethics_approved=True,
                    user_metadata={
                        "my updated custom": "annotation",
                    }
                )
            ]
        )
    ),
    # ENTITY - MODEL
    RouteParameters(
        route="/entity/model",
        name="Entity - Model",
        category=ItemCategory.ENTITY,
        subtype=ItemSubType.MODEL,
        typing_information=TypeParameters[
            ModelDomainInfo,
            ItemModel,
            ModelSeedResponse,
            ModelFetchResponse,
            ModelListResponse,
            ModelCreateResponse
        ](
            domain_info=ModelDomainInfo,
            item_model=ItemModel,
            seed_response=ModelSeedResponse,
            fetch_response=ModelFetchResponse,
            list_response=ModelListResponse,
            create_response=ModelCreateResponse
        ),
        model_examples=ModelExamples[ModelDomainInfo](
            domain_info=[
                ModelDomainInfo(
                    display_name="Example model",
                    name="Example model",
                    description="This is a fake model",
                    documentation_url="https://example_model.org",
                    source_url="https://example_model.org",
                    user_metadata={
                        "my custom": "annotation",
                        "another custom": "annotation"
                    }
                ),
                ModelDomainInfo(
                    display_name="Example model updated",
                    name="Example model updated",
                    description="This is a fake updated model",
                    documentation_url="https://example_model.org",
                    source_url="https://example_model.org",
                    user_metadata={
                        "my updated custom": "annotation",
                    }
                )
            ]
        )
    ),

    # ENTITY - MODEL RUN WORKFLOW DEF'N
    RouteParameters(
        route="/entity/model_run_workflow",
        name="Entity - Model Run Workflow Def'n",
        category=ItemCategory.ENTITY,
        subtype=ItemSubType.MODEL_RUN_WORKFLOW_TEMPLATE,
        typing_information=TypeParameters[
            ModelRunWorkflowTemplateDomainInfo,
            ItemModelRunWorkflowTemplate,
            ModelRunWorkflowTemplateSeedResponse,
            ModelRunWorkflowTemplateFetchResponse,
            ModelRunWorkflowTemplateListResponse,
            ModelRunWorkflowTemplateCreateResponse
        ](
            domain_info=ModelRunWorkflowTemplateDomainInfo,
            item_model=ItemModelRunWorkflowTemplate,
            seed_response=ModelRunWorkflowTemplateSeedResponse,
            fetch_response=ModelRunWorkflowTemplateFetchResponse,
            list_response=ModelRunWorkflowTemplateListResponse,
            create_response=ModelRunWorkflowTemplateCreateResponse
        ),
        model_examples=ModelExamples[ModelRunWorkflowTemplateDomainInfo](
            domain_info=[
                ModelRunWorkflowTemplateDomainInfo(
                    display_name="This is an example workflow def'n for a model run",
                    software_id="1234",
                    input_templates=[
                        TemplateResource(template_id='1234', optional=False)
                    ],
                    output_templates=[
                        TemplateResource(template_id='123456')
                    ],
                    annotations=WorkflowTemplateAnnotations(
                        required=["annotation_key1"],
                        optional=["annotation_key2"]
                    ),
                    user_metadata={
                        "my custom": "annotation",
                        "another custom": "annotation"
                    }
                ),
                ModelRunWorkflowTemplateDomainInfo(
                    display_name="This is an example workflow def'n for a model run",
                    software_id="1234",
                    annotations=WorkflowTemplateAnnotations(
                        required=["annotation_key1"],
                        optional=["annotation_key2"]
                    ),
                    user_metadata={
                        "my updated custom": "annotation",
                    }
                ),
                ModelRunWorkflowTemplateDomainInfo(
                    display_name="This is an updated example workflow def'n for a model run",
                    software_id="1234",
                    input_templates=[
                        TemplateResource(template_id='12345', optional=False)
                    ],
                    output_templates=[
                        TemplateResource(template_id='123456')
                    ],
                    annotations=WorkflowTemplateAnnotations(
                        required=["annotation_key3"],
                        optional=["annotation_key4"]
                    ),
                    user_metadata={
                        "my updated 2 custom": "annotation",
                    }
                ),
            ]
        )
    ),

    # ENTITY - DATASET TEMPLATE
    RouteParameters(
        route="/entity/dataset_template",
        name="Entity - Dataset Template",
        category=ItemCategory.ENTITY,
        subtype=ItemSubType.DATASET_TEMPLATE,
        typing_information=TypeParameters[
            DatasetTemplateDomainInfo,
            ItemDatasetTemplate,
            DatasetTemplateSeedResponse,
            DatasetTemplateFetchResponse,
            DatasetTemplateListResponse,
            DatasetTemplateCreateResponse
        ](
            domain_info=DatasetTemplateDomainInfo,
            item_model=ItemDatasetTemplate,
            seed_response=DatasetTemplateSeedResponse,
            fetch_response=DatasetTemplateFetchResponse,
            list_response=DatasetTemplateListResponse,
            create_response=DatasetTemplateCreateResponse
        ),
        model_examples=ModelExamples[DatasetTemplateDomainInfo](
            domain_info=[
                DatasetTemplateDomainInfo(
                    description="This is an example dataset template",
                    display_name="Example template",
                    defined_resources=[
                        DefinedResource(
                            usage_type=ResourceUsageType.GENERAL_DATA,
                            description="Used for connectivities",
                            path="forcing/",
                        )
                    ],
                    deferred_resources=[
                        DeferredResource(
                            usage_type=ResourceUsageType.GENERAL_DATA,
                            description="Used for connectivities",
                            key="item_key",
                        )
                    ],
                    user_metadata={
                        "my custom": "annotation",
                        "another custom": "annotation"
                    }
                ),
                DatasetTemplateDomainInfo(
                    description="This is an example dataset template",
                    display_name="Example template",
                    user_metadata={
                        "my updated custom": "annotation",
                    }
                ),
                DatasetTemplateDomainInfo(
                    description="This is an example of an updated dataset template",
                    display_name="Example template updated",
                    defined_resources=[
                        DefinedResource(
                            usage_type=ResourceUsageType.GENERAL_DATA,
                            description="Used for connectivities",
                            path="forcing/newpath",
                        )
                    ],
                    deferred_resources=[
                        DeferredResource(
                            usage_type=ResourceUsageType.GENERAL_DATA,
                            description="Used for connectivities",
                            key="item_key",
                        )
                    ],
                    user_metadata={
                        "my updated 2 custom": "annotation",
                    }
                ),
            ]
        )
    ),

    # ACTIVITY - STUDY
    RouteParameters(
        route="/activity/study",
        name="Agent - Study",
        category=ItemCategory.ACTIVITY,
        subtype=ItemSubType.STUDY,
        typing_information=TypeParameters[
            StudyDomainInfo,
            ItemStudy,
            StudySeedResponse,
            StudyFetchResponse,
            StudyListResponse,
            StudyCreateResponse
        ](
            domain_info=StudyDomainInfo,
            item_model=ItemStudy,
            seed_response=StudySeedResponse,
            fetch_response=StudyFetchResponse,
            list_response=StudyListResponse,
            create_response=StudyCreateResponse
        ),
        model_examples=ModelExamples[StudyDomainInfo](
            domain_info=[
                StudyDomainInfo(
                    display_name="Test Study 1",
                    study_alternative_id="1234",
                    title="Test Study 1",
                    description="Test Study 1",
                    user_metadata={
                        "my custom": "annotation",
                        "another custom": "annotation"
                    }
                ),
                StudyDomainInfo(
                    display_name="Test Study 2",
                    title="Test Study 2",
                    description="Test Study 2",
                    user_metadata={
                        "my updated custom": "annotation",
                    }
                )
            ]
        )
    ),

]

# E.g. Create, Version
non_test_route_params: List[RouteParameters] = [
    # ACTIVITY - REGISTRY CREATE
    RouteParameters(
        # This is read only so not really applicable
        use_special_proxy_modify_routes=False,
        route="/activity/create",
        name="Activity - Registry Create",
        category=ItemCategory.ACTIVITY,
        subtype=ItemSubType.CREATE,
        typing_information=TypeParameters[
            CreateDomainInfo,
            ItemCreate,
            None,  # no seed response
            CreateFetchResponse,
            CreateListResponse,
            None,  # no create response
        ](
            domain_info=CreateDomainInfo,
            item_model=ItemCreate,
            seed_response=None,
            fetch_response=CreateFetchResponse,
            list_response=CreateListResponse,
            create_response=None
        ),
        model_examples=ModelExamples[CreateDomainInfo](
            domain_info=[
                CreateDomainInfo(
                    display_name="Example Registry Create Activity Entity",
                    created_item_id="1234",
                    user_metadata={
                        "my custom": "annotation",
                        "another custom": "annotation"
                    }
                ),
                CreateDomainInfo(
                    display_name="Example Registry Create Activity Entity 2 ",
                    created_item_id="5678",
                    user_metadata={
                        "my updated custom": "annotation",
                    }
                )
            ]

        )
    ),

    # ACTIVITY - REGISTRY VERSION
    RouteParameters(
        # This is read only so not really applicable
        use_special_proxy_modify_routes=False,
        route="/activity/version",
        name="Activity - Registry Version",
        category=ItemCategory.ACTIVITY,
        subtype=ItemSubType.VERSION,
        typing_information=TypeParameters[
            VersionDomainInfo,
            ItemVersion,
            None,  # no seed response
            VersionFetchResponse,
            VersionListResponse,
            None,  # no create response
        ](
            domain_info=VersionDomainInfo,
            item_model=ItemVersion,
            seed_response=None,
            fetch_response=VersionFetchResponse,
            list_response=VersionListResponse,
            create_response=None
        ),
        model_examples=ModelExamples[VersionDomainInfo](
            domain_info=[
                VersionDomainInfo(
                    display_name="Example Registry Version Activity Entity 1",
                    from_item_id="123",
                    to_item_id="1234",
                    new_version_number=2,
                    reason="test",
                    user_metadata={
                        "my custom": "annotation",
                        "another custom": "annotation"
                    }
                ),
                VersionDomainInfo(
                    display_name="Example Registry Version Activity Entity 2",
                    from_item_id="1234",
                    to_item_id="12345",
                    new_version_number=3,
                    reason="test 2",
                    user_metadata={
                        "my updated custom": "annotation",
                    }
                )
            ]

        )
    ),
]



from setuptools import setup, find_packages  # type: ignore
import os
from package_readme import long_description

# 0.1.0 is the 'version' installs from this directory from other components in repo will use
version = os.getenv('TAG_NAME', 'v0.1.0')

setup(
    name='provena_interfaces_v2',
    # format acceptably as 'v0.0.1' or '0.0.1',
    version=version,
    description='Interfaces for Provena Application (see https://provena.io)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(
        include=[
            'ProvenaInterfaces',
            'ProvenaInterfaces.*',
            'ProvenaInterfaces.*.*',
        ]
    ),
    install_requires=[
        'fastapi==0.116.1',
        'pydantic>=2.0',
        'email-validator',
        'isodate',
    ],
    package_data={
        'ProvenaInterfaces': ['py.typed']
    },
    url="https://provena.io",
    maintainer_email="rrap-mds-is-support@csiro.au",

)



from __future__ import annotations
from typing import Dict, Any, List, Optional, Tuple, Union, ForwardRef
from pydantic import ConfigDict, BaseModel, Field, HttpUrl, AnyHttpUrl, validator
from datetime import datetime, date
from enum import Enum

HOURS_TO_SECS = 3600


# Interface exports prefer relative imports
# where as pip install prefers module level import
try:
    from ProvenaInterfaces.SharedTypes import Status, StatusResponse
    from ProvenaInterfaces.RegistryModels import ItemDataset, S3Location, CollectionFormat, Roles, IdentifiedResource
except:
    from .SharedTypes import Status, StatusResponse
    from .RegistryModels import ItemDataset, S3Location, CollectionFormat, Roles, IdentifiedResource

Handle = str
ROCrate = Dict[str, Any]

# RONode = ForwardRef('RONode')


class ROConnectionFlat(BaseModel):
    name: str
    value: Union[str, List[str]]
    model_config = ConfigDict(str_min_length=1, extra="forbid")


class ROConnectionIDNonExpanded(BaseModel):
    name: str
    value_id: str
    model_config = ConfigDict(str_min_length=1, extra="forbid")


class ROConnectionExpandedUniqueName(BaseModel):
    name: str
    to: RONode
    model_config = ConfigDict(str_min_length=1, extra="forbid")


class ROConnectionExpandedIdName(BaseModel):
    name: str
    to: RONode
    special_id: str
    model_config = ConfigDict(str_min_length=1, extra="forbid")


ROConnectionType = Union[ROConnectionFlat, ROConnectionIDNonExpanded,
                         ROConnectionExpandedUniqueName, ROConnectionExpandedIdName]


class RONode(BaseModel):
    connections: Dict[str, ROConnectionType]
    type: Union[str, List[str]]
    model_config = ConfigDict(str_min_length=1, extra="forbid")


ROConnectionExpandedUniqueName.update_forward_refs()
ROConnectionExpandedIdName.update_forward_refs()


class ROGraph(BaseModel):
    root_node: RONode
    model_config = ConfigDict(str_min_length=1, extra="forbid")


class Schema(BaseModel):
    json_schema: Dict[Any, Any]


class MintResponse(BaseModel):
    status: Status
    handle: Optional[str] = None
    s3_location: Optional[S3Location] = None
    register_create_activity_session_id: Optional[str] = None


class UpdateMetadataResponse(BaseModel):
    status: Status
    handle: str
    s3_location: S3Location

class Credentials(BaseModel):
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_session_token: str
    expiry: datetime


class CredentialsRequest(BaseModel):
    dataset_id: str
    console_session_required: bool


class CredentialResponse(BaseModel):
    status: Status
    credentials: Credentials
    console_session_url: Optional[str] = None


# class LockActionType(str, Enum):
#    LOCK = "LOCK"
#    UNLOCK = "UNLOCK"
#
#
# class LockEvent(BaseModel):
#    # Locked or unlocked
#    action_type: LockActionType
#    # Who did it?
#    actor_email: str
#    # Why did they lock/unlock it
#    reason: str
#    # When did this happen? (unix timestamp)
#    timestamp: int
#
#
# class DatasetLockConfiguration(BaseModel):
#    # is the dataset locked down?
#    locked: bool
#    # what is the history of lock changes
#    history: List[LockEvent]


class RegistryFetchResponse(StatusResponse):
    item: Optional[ItemDataset] = None
    roles: Optional[Roles] = None
    locked: Optional[bool] = None


class ListRegistryResponse(BaseModel):
    status: Status
    num_items: int
    registry_items: List[ItemDataset]
    pagination_key: Optional[Dict[str, Any]] = None


class InputConvertRocrateCollectionFormat(BaseModel):
    rocrate_items: List[Dict[str, Any]]


class CollectionFormatConversion(BaseModel):
    handle: str
    collection_format: CollectionFormat
    s3: S3Location
    created_time: datetime
    updated_time: datetime


class InputConvertCollectionFormatRocrate(BaseModel):
    collection_format_items: List[CollectionFormatConversion]


class PossibleCollectionFormat(BaseModel):
    success: bool
    collection_format: Optional[CollectionFormat] = None
    error_message: Optional[str] = None


class PossibleRocrate(BaseModel):
    success: bool
    rocrate: Optional[ROCrate] = None
    error_message: Optional[str] = None


class ResponseConvertRocrateCollectionFormat(BaseModel):
    collection_format_items: List[PossibleCollectionFormat]


class ResponseConvertCollectionFormatRocrate(BaseModel):
    rocrate_items: List[PossibleRocrate]


class ReleaseApprovalRequest(BaseModel):
    dataset_id: IdentifiedResource
    approver_id: IdentifiedResource
    notes: str

class ReleaseApprovalRequestResponse(BaseModel):
    dataset_id: IdentifiedResource
    approver_id: IdentifiedResource
    details: str


class ActionApprovalRequest(BaseModel):
    dataset_id: IdentifiedResource
    approve: bool
    notes: str

class ActionApprovalRequestResponse(BaseModel):
    dataset_id: IdentifiedResource
    approved: bool
    details: str


class PresignedURLRequest(BaseModel):
    dataset_id: IdentifiedResource
    file_path: str
    expires_in: int = Field(
        HOURS_TO_SECS*3, 
        description="The number of seconds the presigned URL is valid for. Defaults to 3 hours (3600*3).",
        ge=1,
        le=HOURS_TO_SECS*24,
    )

class PresignedURLResponse(BaseModel):
    dataset_id: IdentifiedResource
    file_path: str
    presigned_url: str